;; testbild.asm                                      letzte Aenderung: 22.6.2014
;* Variante fuer RGB-Ausgabe
;*
;* Schaltschema: basirgb.png
;*               zum Debuggen zusaetzlich LEDs an PB0 und PB3
;*   Eventuell einfachere Verdrahtung falls Fernseher das unterstuetzt:
;*   Schaltpegel an Scart-Pin16 u. Syncsignal Pin20 weglassen, dafuer
;*   Sync zum Gruen zumischen (1k + Diode, statt R2 330 Ohm + Diode)
;*
;* Letzte Aenderungen:
;* 11.6.2014   Version 0.1, Erstellung
;*
;* Copyright: Opensource, Freeware
;*
;* VERSION.REVISION = "0.1"
#define VERSION 0
#define REVISION 1
;*
;-----------------------------------------------------------------------------
;- Useroptionen:

.equ RAND_LINKS=1700	  ;Linker Rand in Nanosekunden (Default 1700)
.equ RAND_OBEN=4	  ;Oberer Rand in Zeilen (optimale Einstellung abhaengig vom verwendeten Monitor
			  ;2 fuer Analogfernseher, 12 fuer Digitalfernseher)

.equ HFARBE=0x40	  ;Hintergrund-Farbe (0x70=weiss, 0x00=schwarz, 0x10 rot, 0x20 gruen, 0x40 blau)
.equ TFARBE=0x20	  ;Text-Farbe oder Vordergrundfarbe

;#define LOWRES		;Bild mit kleiner Aufloesung: sollte Groesse 160*120 haben (X*Y)
#define MEDRES		;Bild mit mittlerer horizontaler Aufloesung: sollte Groesse 192*132 haben (X*Y)
;#define HIWRES		;Bild mit bester horizontaler Aufloesung: sollte Groesse 240*264 haben (X*Y) (zu gross fuer RAM)
			;(bisher nur MEDRES getestet)

;#define INTERLACE	;fuer Interlace-Modus (noch nicht programmiert)

#define ATMEGA1284P	;bisher nur mit diesem getestet
;#define ATMEGA328P	;mit diesem koennte es auch gehen (noch nicht getestet)
.equ F_CPU=20000000	;braucht immer 20 MHz

;- Ende Useroptionen
;-----------------------------------------------------------------------------
#define DEBUG		  ;zur Fehlersuche
	
.include "ifndeftest.inc"	;test damit man vor veraltetem avra gewarnt wird

.equ RAND_RECHTS=4000-RAND_LINKS	;Rechter Rand in Nanosekunden (Default 2300)
.equ oberedunkelzeilen=26 ;(von Monitor abhaengig? 26 fuer Analogfernseher, 36 fuer Digitalfernseher bei RAND_OBEN=2)

;; Eine normale Video-Zeile sieht so aus:
;; Schwarzschulter   Sync-Puls     Schwarzschulter, 
;; 1.5us (vordere)   4.7us HSync,  5.8us (hintere)   1.7us Rand links, 48us Text, 2.3us Rand rechts --> total 64us
;; 30 Takte          94 Takte      116 Takte         34 Takte          960 Takte  46      --> 1280 Takte bei 20MHz

.equ MU=F_CPU/1000000	;Ticks fuer 1 Microsekunde
.equ VORDERE_SCHWARZSCHULTER=MU*15/10
.equ SYNC1=MU*47/10		;4.7us fuer HSYNC, in Ticks
.equ SYNC2=140		;Anzahl Microsekunden fuer VSYNC-Puls mit Interlace (140-160, mind 132, max 180)
.equ SYNC3=160		;Anzahl Microsekunden fuer VSYNC-Puls ohne Interlace
.equ HINTERE_SCHWARZSCHULTER=12*MU-VORDERE_SCHWARZSCHULTER-SYNC1  ; Hinter = 12us - Vordere - Sync1
.equ RANDLINKS=RAND_LINKS*MU/1000	;Umrechnung von ns in Ticks (ergibt 40 bei Defaultwert)
.equ LEER1=HINTERE_SCHWARZSCHULTER+RANDLINKS
.equ RANDRECHTS=RAND_RECHTS*MU/1000	;Umrechnung von ns in Ticks (ergibt 40 bei Defaultwert)
.equ INHALT=52*MU-RANDLINKS-RANDRECHTS	;Inhalt in horizontaler Richtung in Ticks (48us = 960 bei 20MHz)

#if INHALT!=(48*MU)
	.warning "INHALT ist ungleich 48 Microsekunden" ;test
#endif

.equ Z312=312
.equ zeilentotal=Z312
.equ INHALTZEILEN=264
.equ obereleerzeilen=oberedunkelzeilen+RAND_OBEN
.equ untereleerzeilen=zeilentotal-obereleerzeilen-INHALTZEILEN

.if untereleerzeilen<2
	.warning "Oberer Rand zu gross"
	.message "RAND_OBEN und Dunkelzeilen zusammen maximal 46"
.endif

;-----------------------------------------------------------------------------
;; Beginn des eigentlichen Codes

#ifdef ATMEGA328P
.include "m328Pdef.inc"		;ATmega328P
#endif
#ifdef ATMEGA1284P
.include "m1284Pdef.inc"	;ATmega1284P
#endif

	jmp main
.org OC1Aaddr
	jmp timer1_match_A

#ifdef INTERLACE
.org OC1Baddr
	jmp timer1_match_B
#endif

;-----------------------------------------------------------------------------
; Globale Registerbelegungen:
; r4=Sicherung von SREG im Interrupt
; r5=0  Global auf 0 gesetzt
; r7:r6 = Zwischenspeicher fuer Y
; r3:r2 = Zeiger ins Flash auf das Testbild
; r15 Widerhohlungszaehler fuer Videozeilen
.def zero=r5
.def zeilenzaehlerL=r24
.def zeilenzaehlerH=r25
.def syncstatus=r8
#ifdef INTERLACE
.def halbbildflag=r9
.def mitteflag=r10
.def vsyncflag=r11
#endif
.def hintergrundfarbe=r22
.def textfarbe=r23
; Noch im Hauptprogramm verwendbar:
; r0-r1, r12-r14, r16-r21
; XH:XL (r27:r26)
; YH:YL (r29:r28)
; ZH:ZL (r31:r30)

;-----------------------------------------------------------------------------
; Registerbelegung im Hauptprogramm:
.def status=r19
.def altsec=r14

;-----------------------------------------------------------------------------
; Konstanten:
  .equ SYNC_PORT=PORTB
  .equ SYNC_DDR =DDRB
  .equ SYNCBIT=PB1
  .equ BILD_PORT=PORTA
  .equ BILD_DDR =DDRA
  .equ BILDBITR=PA4
  .equ BILDBITG=PA5
  .equ BILDBITB=PA6

#ifdef DEBUG
  .equ LED_PORT=PORTB
  .equ LED_DDR=DDRB
  .equ LED1=0
  .equ LED2=3
#endif

.equ SERBUFFERSIZE=8		;Fuer rs232-Empfang (nur Werte 2^n verwendbar)

;-----------------------------------------------------------------------------
; Macros:

.macro	ticksw
.endm
.macro	ticksw_i
  #if @0<0
	.warning "negativer Wert in macro ticksw"
  #endif
  #if @0==1
	nop
  #endif
  #if @0==2
	adiw r24,0
  #endif
  #if @0==3
	nop
	adiw r24,0
  #endif
  #if @0==4
	adiw r24,0
	adiw r24,0
  #endif
  #if @0==5
	nop
	adiw r24,0
	adiw r24,0
  #endif
  #if @0==6
	adiw r24,0
	adiw r24,0
	adiw r24,0
  #endif
  #if @0==7
	rcall tickswait7
  #endif
  #if @0==8
	rcall tickswait8
  #endif
  #if @0==9
	rcall tickswait9
  #endif
  #if @0==10
	rcall tickswait10
  #endif
  #if @0==11
	rcall tickswait11
  #endif
  #if @0==12
	rcall tickswait12
  #endif
  #if @0==13
	rcall tickswait13
  #endif
  #if @0==14
	rcall tickswait14
  #endif
  #if @0==15
	rcall tickswait15
  #endif
  #if @0==16
	rcall tickswait16
  #endif
  #if @0==17
	rcall tickswait17
  #endif
  #if @0>=18
	ldi r16, @0
	rcall tickswait
  #endif
.endm

.macro input
.endm
.macro input_8_i
#if @1>=64
	lds @0, @1
#else
	in  @0, @1
#endif
.endm

.macro output
.endm
.macro output_i_8
#if @0>=64
	sts @0, @1
#else
	out @0, @1
#endif
.endm

.equ MINA=14	;dies entspricht der minimalen Anzahl Takte bis TCNT1L eingelesen wird
.equ MAXA=20	;Maximaler Wert (6 mehr als MINA bei Verwendung von sleep, sonst sollte 4 reichen)

.macro jitterausgleich
	lds ZL, TCNT1L		;2 Takte
#ifdef JITTER_MINMAXTEST
	cp ZL, minwert
	brcc ji1
	mov minwert, ZL		;Minimalwert speichern
  ji1:	cp ZL, maxwert
	brcs ji2
	mov maxwert, ZL		;Maximalwert speichern
  ji2:
#endif
	clr  ZH
	cpi ZL, MAXA+1		;TCNT1 doch groesser als Maximaler Zaehlerstand
	brcc fehler		;ja--> Absturz verhindern
	subi ZL, low(-(L1-MINA))  ;16-Bit-Addition von (L1-Minimum)
	sbci ZH, high(-(L1-MINA)) ;also Z = L1 + Zaehlerstand-Minimum
	ijmp
fehler:	sbi LED_PORT, LED2	;Fehlertest
	nop
	nop
	nop
	nop	;zusaetzliche nop falls TCNT1 doch kleiner als MINA ist
	sbi LED_PORT, LED1	;Fehlertest
L1:	nop	;bis zu 6 Takte fuer Jitterausgleich
	nop
	nop
	nop
	nop
	nop
	; 9 bis 15 Takte gebraucht (6 mehr mit MINMAXTEST)
	;; hier sollte es jitterfrei sein.
.endm

#ifdef JITTER_MINMAXTEST
.set JITAKTE=21
#else
.set JITAKTE=15
#endif
	
;-----------------------------------------------------------------------------
; Variablen im SRAM:
                .DSEG
serbuffer:	.BYTE SERBUFFERSIZE
jread:		.BYTE 1			; wenn jread==jwrite dann ist der Buffer leer
jwrite:		.BYTE 1			; wenn jread==jwrite+1 dann ist der Buffer voll
;textspeicher:	.BYTE 2*ANZAHLTEXTBYTES ; Zeichen+Farbe braucht 2 Bytes
uhr_ticks:	.BYTE 1		; 50stels Sekunden (Wertebereich 0...49)
uhr_sec:	.BYTE 1		; Zeit seit dem Einschalten
uhr_min:	.BYTE 1
uhr_std:	.BYTE 1
uhr_tage:	.BYTE 1
pieps_endzeit:	.BYTE 1		; Endzeit+1 fuer Piepser in Ticks (+1 fuer Wertebereich 1...50)
		.CSEG

;-----------------------------------------------------------------------------
;; Testbild im FLASH:
;; Bild mit kleinerer Aufloesung: 192 breit, 132 hoch
;; Bild mit groesserer Aufloesung: 240 breit, 264 hoch
testbild:
#ifdef MEDRES

;; Bild mit mittlerer Aufloesung: 192 breit, 132 hoch
;; 1. Zeile:
.db 0xFF,0xFF,0xFF,0xFF,0xFF,0xFF,0xFF,0xFF,0xFF,0xFF	;erste 10 Bytes = 20 Pixel, weiss
.db 0xFF,0xFF,0xFF,0xFF,0xFF,0xFF,0xFF,0xFF,0xFF,0xFF	; - 40. Pixel, weiss
.db 0xFF,0xFF,0xFF,0xFF,0xFF,0xFF,0xFF,0xFF,0xFF,0xFF	; - 60. Pixel, weiss
.db 0xFF,0xFF,0xFF,0xFF,0xFF,0xFF,0xFF,0xFF,0xFF,0xFF	; - 80. Pixel, weiss
.db 0xFF,0xFF,0xFF,0xFF,0xFF,0xFF,0xFF,0xFF,0xFF,0xFF	; - 100. Pixel, weiss
.db 0xFF,0xFF,0xFF,0xFF,0xFF,0xFF,0xFF,0xFF,0xFF,0xFF	; - 120. Pixel, weiss
.db 0xFF,0xFF,0xFF,0xFF,0xFF,0xFF,0xFF,0xFF,0xFF,0xFF	; - 140. Pixel, weiss
.db 0xFF,0xFF,0xFF,0xFF,0xFF,0xFF,0xFF,0xFF,0xFF,0xFF	; - 160. Pixel, weiss
.db 0xFF,0xFF,0xFF,0xFF,0xFF,0xFF,0xFF,0xFF,0xFF,0xFF	; - 180. Pixel, weiss
.db 0xFF,0xFF,0xFF,0xFF,0xFF,0xFF			;181.-192. Pixel, weiss
;; 2. Zeile:
.db 0xF0,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00	;erste 10 Bytes = 20 Pixel, erstes Pixel weiss, rest schwarz
.db 0x11,0x11,0x11,0x11,0x11,0x11,0x11,0x11,0x11,0x11	;21.-40. Pixel, rot
.db 0x22,0x22,0x22,0x22,0x22,0x22,0x22,0x22,0x22,0x22	;41.-60. Pixel, gruen
.db 0x44,0x44,0x44,0x44,0x44,0x44,0x44,0x44,0x44,0x44	;61.-80. Pixel, blau
.db 0x77,0x77,0x77,0x77,0x77,0x77,0x77,0x77,0x77,0x77	;21.-100. Pixel, weiss
.db 0x33,0x33,0x33,0x33,0x33,0x33,0x33,0x33,0x33,0x33	;21.-120. Pixel, gelb
.db 0x55,0x55,0x55,0x55,0x55,0x55,0x55,0x55,0x55,0x55	;21.-140. Pixel, cyan
.db 0x66,0x66,0x66,0x66,0x66,0x66,0x66,0x66,0x66,0x66	;21.-160. Pixel, rotblau
.db 0x70,0x70,0x70,0x70,0x70,0x70,0x70,0x70,0x70,0x70	;21.-180. Pixel, grau
.db 0x00,0x00,0x00,0x00,0x00,0x0F			;181.-192. Pixel, schwarz, letztes Pixel weiss
;; 3. Zeile:					
.db 0xF0,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00	;erste 10 Bytes = 20 Pixel, schwarz
.db 0x11,0x11,0x11,0x11,0x11,0x11,0x11,0x11,0x11,0x11	;21.-40. Pixel, rot
.db 0x22,0x22,0x22,0x22,0x22,0x22,0x22,0x22,0x22,0x22	;41.-60. Pixel, gruen
.db 0x44,0x44,0x44,0x44,0x44,0x44,0x44,0x44,0x44,0x44	;61.-80. Pixel, blau
.db 0x77,0x77,0x77,0x77,0x77,0x77,0x77,0x77,0x77,0x77	;21.-100. Pixel, weiss
.db 0x33,0x33,0x33,0x33,0x33,0x33,0x33,0x33,0x33,0x33	;21.-120. Pixel, gelb
.db 0x55,0x55,0x55,0x55,0x55,0x55,0x55,0x55,0x55,0x55	;21.-140. Pixel, cyan
.db 0x66,0x66,0x66,0x66,0x66,0x66,0x66,0x66,0x66,0x66	;21.-160. Pixel, rotblau
.db 0x70,0x70,0x70,0x70,0x70,0x70,0x70,0x70,0x70,0x70	;21.-180. Pixel, grau
.db  0x00,0x00,0x00,0x00,0x00,0x0F			;181.-192. Pixel, schwarz
;; 4. Zeile:					
.db 0xF0,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00	;erste 10 Bytes = 20 Pixel, schwarz
.db 0x11,0x11,0x11,0x11,0x11,0x11,0x11,0x11,0x11,0x11	;21.-40. Pixel, rot
.db 0x22,0x22,0x22,0x22,0x22,0x22,0x22,0x22,0x22,0x22	;41.-60. Pixel, gruen
.db 0x44,0x44,0x44,0x44,0x44,0x44,0x44,0x44,0x44,0x44	;61.-80. Pixel, blau
.db 0x77,0x77,0x77,0x77,0x77,0x77,0x77,0x77,0x77,0x77	;21.-100. Pixel, weiss
.db 0x33,0x33,0x33,0x33,0x33,0x33,0x33,0x33,0x33,0x33	;21.-120. Pixel, gelb
.db 0x55,0x55,0x55,0x55,0x55,0x55,0x55,0x55,0x55,0x55	;21.-140. Pixel, cyan
.db 0x66,0x66,0x66,0x66,0x66,0x66,0x66,0x66,0x66,0x66	;21.-160. Pixel, rotblau
.db 0x70,0x70,0x70,0x70,0x70,0x70,0x70,0x70,0x70,0x70	;21.-180. Pixel, grau
.db  0x00,0x00,0x00,0x00,0x00,0x0F			;181.-192. Pixel, schwarz
;; 5. Zeile:					
.db 0xF0,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00	;erste 10 Bytes = 20 Pixel, schwarz
.db 0x11,0x11,0x11,0x11,0x11,0x11,0x11,0x11,0x11,0x11	;21.-40. Pixel, rot
.db 0x22,0x22,0x22,0x22,0x22,0x22,0x22,0x22,0x22,0x22	;41.-60. Pixel, gruen
.db 0x44,0x44,0x44,0x44,0x44,0x44,0x44,0x44,0x44,0x44	;61.-80. Pixel, blau
.db 0x77,0x77,0x77,0x77,0x77,0x77,0x77,0x77,0x77,0x77	;21.-100. Pixel, weiss
.db 0x33,0x33,0x33,0x33,0x33,0x33,0x33,0x33,0x33,0x33	;21.-120. Pixel, gelb
.db 0x55,0x55,0x55,0x55,0x55,0x55,0x55,0x55,0x55,0x55	;21.-140. Pixel, cyan
.db 0x66,0x66,0x66,0x66,0x66,0x66,0x66,0x66,0x66,0x66	;21.-160. Pixel, rotblau
.db 0x70,0x70,0x70,0x70,0x70,0x70,0x70,0x70,0x70,0x70	;21.-180. Pixel, grau
.db  0x00,0x00,0x00,0x00,0x00,0x0F			;181.-192. Pixel, schwarz
;; 6. Zeile:					
.db 0xF0,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00	;erste 10 Bytes = 20 Pixel, schwarz
.db 0x11,0x11,0x11,0x11,0x11,0x11,0x11,0x11,0x11,0x11	;21.-40. Pixel, rot
.db 0x22,0x22,0x22,0x22,0x22,0x22,0x22,0x22,0x22,0x22	;41.-60. Pixel, gruen
.db 0x44,0x44,0x44,0x44,0x44,0x44,0x44,0x44,0x44,0x44	;61.-80. Pixel, blau
.db 0x77,0x77,0x77,0x77,0x77,0x77,0x77,0x77,0x77,0x77	;21.-100. Pixel, weiss
.db 0x33,0x33,0x33,0x33,0x33,0x33,0x33,0x33,0x33,0x33	;21.-120. Pixel, gelb
.db 0x55,0x55,0x55,0x55,0x55,0x55,0x55,0x55,0x55,0x55	;21.-140. Pixel, cyan
.db 0x66,0x66,0x66,0x66,0x66,0x66,0x66,0x66,0x66,0x66	;21.-160. Pixel, rotblau
.db 0x70,0x70,0x70,0x70,0x70,0x70,0x70,0x70,0x70,0x70	;21.-180. Pixel, grau
.db  0x00,0x00,0x00,0x00,0x00,0x0F			;181.-192. Pixel, schwarz
;; 7. Zeile:					
.db 0xF0,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00	;erste 10 Bytes = 20 Pixel, schwarz
.db 0x11,0x11,0x11,0x11,0x11,0x11,0x11,0x11,0x11,0x11	;21.-40. Pixel, rot
.db 0x22,0x22,0x22,0x22,0x22,0x22,0x22,0x22,0x22,0x22	;41.-60. Pixel, gruen
.db 0x44,0x44,0x44,0x44,0x44,0x44,0x44,0x44,0x44,0x44	;61.-80. Pixel, blau
.db 0x77,0x77,0x77,0x77,0x77,0x77,0x77,0x77,0x77,0x77	;21.-100. Pixel, weiss
.db 0x33,0x33,0x33,0x33,0x33,0x33,0x33,0x33,0x33,0x33	;21.-120. Pixel, gelb
.db 0x55,0x55,0x55,0x55,0x55,0x55,0x55,0x55,0x55,0x55	;21.-140. Pixel, cyan
.db 0x66,0x66,0x66,0x66,0x66,0x66,0x66,0x66,0x66,0x66	;21.-160. Pixel, rotblau
.db 0x70,0x70,0x70,0x70,0x70,0x70,0x70,0x70,0x70,0x70	;21.-180. Pixel, grau
.db  0x00,0x00,0x00,0x00,0x00,0x0F			;181.-192. Pixel, schwarz
;; 8. Zeile:					
.db 0xF0,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00	;erste 10 Bytes = 20 Pixel, schwarz
.db 0x11,0x11,0x11,0x11,0x11,0x11,0x11,0x11,0x11,0x11	;21.-40. Pixel, rot
.db 0x22,0x22,0x22,0x22,0x22,0x22,0x22,0x22,0x22,0x22	;41.-60. Pixel, gruen
.db 0x44,0x44,0x44,0x44,0x44,0x44,0x44,0x44,0x44,0x44	;61.-80. Pixel, blau
.db 0x77,0x77,0x77,0x77,0x77,0x77,0x77,0x77,0x77,0x77	;21.-100. Pixel, weiss
.db 0x33,0x33,0x33,0x33,0x33,0x33,0x33,0x33,0x33,0x33	;21.-120. Pixel, gelb
.db 0x55,0x55,0x55,0x55,0x55,0x55,0x55,0x55,0x55,0x55	;21.-140. Pixel, cyan
.db 0x66,0x66,0x66,0x66,0x66,0x66,0x66,0x66,0x66,0x66	;21.-160. Pixel, rotblau
.db 0x70,0x70,0x70,0x70,0x70,0x70,0x70,0x70,0x70,0x70	;21.-180. Pixel, grau
.db  0x00,0x00,0x00,0x00,0x00,0x0F			;181.-192. Pixel, schwarz
;; 9. Zeile:					
.db 0xF0,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00	;erste 10 Bytes = 20 Pixel, schwarz
.db 0x11,0x11,0x11,0x11,0x11,0x11,0x11,0x11,0x11,0x11	;21.-40. Pixel, rot
.db 0x22,0x22,0x22,0x22,0x22,0x22,0x22,0x22,0x22,0x22	;41.-60. Pixel, gruen
.db 0x44,0x44,0x44,0x44,0x44,0x44,0x44,0x44,0x44,0x44	;61.-80. Pixel, blau
.db 0x77,0x77,0x77,0x77,0x77,0x77,0x77,0x77,0x77,0x77	;21.-100. Pixel, weiss
.db 0x33,0x33,0x33,0x33,0x33,0x33,0x33,0x33,0x33,0x33	;21.-120. Pixel, gelb
.db 0x55,0x55,0x55,0x55,0x55,0x55,0x55,0x55,0x55,0x55	;21.-140. Pixel, cyan
.db 0x66,0x66,0x66,0x66,0x66,0x66,0x66,0x66,0x66,0x66	;21.-160. Pixel, rotblau
.db 0x70,0x70,0x70,0x70,0x70,0x70,0x70,0x70,0x70,0x70	;21.-180. Pixel, grau
.db  0x00,0x00,0x00,0x00,0x00,0x0F			;181.-192. Pixel, schwarz
;; 10. Zeile:					
.db 0xF0,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00	;erste 10 Bytes = 20 Pixel, schwarz
.db 0x11,0x11,0x11,0x11,0x11,0x11,0x11,0x11,0x11,0x11	;21.-40. Pixel, rot
.db 0x22,0x22,0x22,0x22,0x22,0x22,0x22,0x22,0x22,0x22	;41.-60. Pixel, gruen
.db 0x44,0x44,0x44,0x44,0x44,0x44,0x44,0x44,0x44,0x44	;61.-80. Pixel, blau
.db 0x77,0x77,0x77,0x77,0x77,0x77,0x77,0x77,0x77,0x77	;21.-100. Pixel, weiss
.db 0x33,0x33,0x33,0x33,0x33,0x33,0x33,0x33,0x33,0x33	;21.-120. Pixel, gelb
.db 0x55,0x55,0x55,0x55,0x55,0x55,0x55,0x55,0x55,0x55	;21.-140. Pixel, cyan
.db 0x66,0x66,0x66,0x66,0x66,0x66,0x66,0x66,0x66,0x66	;21.-160. Pixel, rotblau
.db 0x70,0x70,0x70,0x70,0x70,0x70,0x70,0x70,0x70,0x70	;21.-180. Pixel, grau
.db  0x00,0x00,0x00,0x00,0x00,0x0F			;181.-192. Pixel, schwarz
;; 11.-20. Zeile:
.db 0xF0,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00	;erste 10 Bytes = 20 Pixel, schwarz
.db 0x11,0x11,0x11,0x11,0x11,0x11,0x11,0x11,0x11,0x11	;21.-40. Pixel, rot
.db 0x22,0x22,0x22,0x22,0x22,0x22,0x22,0x22,0x22,0x22	;41.-60. Pixel, gruen
.db 0x44,0x44,0x44,0x44,0x44,0x44,0x44,0x44,0x44,0x44	;61.-80. Pixel, blau
.db 0x77,0x77,0x77,0x77,0x77,0x77,0x77,0x77,0x77,0x77	;21.-100. Pixel, weiss
.db 0x33,0x33,0x33,0x33,0x33,0x33,0x33,0x33,0x33,0x33	;21.-120. Pixel, gelb
.db 0x55,0x55,0x55,0x55,0x55,0x55,0x55,0x55,0x55,0x55	;21.-140. Pixel, cyan
.db 0x66,0x66,0x66,0x66,0x66,0x66,0x66,0x66,0x66,0x66	;21.-160. Pixel, rotblau
.db 0x70,0x70,0x70,0x70,0x70,0x70,0x70,0x70,0x70,0x70	;21.-180. Pixel, grau
.db  0x00,0x00,0x00,0x00,0x00,0x0F			;181.-192. Pixel, schwarz
;; 12. Zeile:					
.db 0xF0,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00	;erste 10 Bytes = 20 Pixel, schwarz
.db 0x11,0x11,0x11,0x11,0x11,0x11,0x11,0x11,0x11,0x11	;21.-40. Pixel, rot
.db 0x22,0x22,0x22,0x22,0x22,0x22,0x22,0x22,0x22,0x22	;41.-60. Pixel, gruen
.db 0x44,0x44,0x44,0x44,0x44,0x44,0x44,0x44,0x44,0x44	;61.-80. Pixel, blau
.db 0x77,0x77,0x77,0x77,0x77,0x77,0x77,0x77,0x77,0x77	;21.-100. Pixel, weiss
.db 0x33,0x33,0x33,0x33,0x33,0x33,0x33,0x33,0x33,0x33	;21.-120. Pixel, gelb
.db 0x55,0x55,0x55,0x55,0x55,0x55,0x55,0x55,0x55,0x55	;21.-140. Pixel, cyan
.db 0x66,0x66,0x66,0x66,0x66,0x66,0x66,0x66,0x66,0x66	;21.-160. Pixel, rotblau
.db 0x70,0x70,0x70,0x70,0x70,0x70,0x70,0x70,0x70,0x70	;21.-180. Pixel, grau
.db  0x00,0x00,0x00,0x00,0x00,0x0F			;181.-192. Pixel, schwarz
;; 13. Zeile:					
.db 0xF0,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00	;erste 10 Bytes = 20 Pixel, schwarz
.db 0x11,0x11,0x11,0x11,0x11,0x11,0x11,0x11,0x11,0x11	;21.-40. Pixel, rot
.db 0x22,0x22,0x22,0x22,0x22,0x22,0x22,0x22,0x22,0x22	;41.-60. Pixel, gruen
.db 0x44,0x44,0x44,0x44,0x44,0x44,0x44,0x44,0x44,0x44	;61.-80. Pixel, blau
.db 0x77,0x77,0x77,0x77,0x77,0x77,0x77,0x77,0x77,0x77	;21.-100. Pixel, weiss
.db 0x33,0x33,0x33,0x33,0x33,0x33,0x33,0x33,0x33,0x33	;21.-120. Pixel, gelb
.db 0x55,0x55,0x55,0x55,0x55,0x55,0x55,0x55,0x55,0x55	;21.-140. Pixel, cyan
.db 0x66,0x66,0x66,0x66,0x66,0x66,0x66,0x66,0x66,0x66	;21.-160. Pixel, rotblau
.db 0x70,0x70,0x70,0x70,0x70,0x70,0x70,0x70,0x70,0x70	;21.-180. Pixel, grau
.db  0x00,0x00,0x00,0x00,0x00,0x0F			;181.-192. Pixel, schwarz
;; 14. Zeile:					
.db 0xF0,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00	;erste 10 Bytes = 20 Pixel, schwarz
.db 0x11,0x11,0x11,0x11,0x11,0x11,0x11,0x11,0x11,0x11	;21.-40. Pixel, rot
.db 0x22,0x22,0x22,0x22,0x22,0x22,0x22,0x22,0x22,0x22	;41.-60. Pixel, gruen
.db 0x44,0x44,0x44,0x44,0x44,0x44,0x44,0x44,0x44,0x44	;61.-80. Pixel, blau
.db 0x77,0x77,0x77,0x77,0x77,0x77,0x77,0x77,0x77,0x77	;21.-100. Pixel, weiss
.db 0x33,0x33,0x33,0x33,0x33,0x33,0x33,0x33,0x33,0x33	;21.-120. Pixel, gelb
.db 0x55,0x55,0x55,0x55,0x55,0x55,0x55,0x55,0x55,0x55	;21.-140. Pixel, cyan
.db 0x66,0x66,0x66,0x66,0x66,0x66,0x66,0x66,0x66,0x66	;21.-160. Pixel, rotblau
.db 0x70,0x70,0x70,0x70,0x70,0x70,0x70,0x70,0x70,0x70	;21.-180. Pixel, grau
.db  0x00,0x00,0x00,0x00,0x00,0x0F			;181.-192. Pixel, schwarz
;; 15. Zeile:					
.db 0xF0,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00	;erste 10 Bytes = 20 Pixel, schwarz
.db 0x11,0x11,0x11,0x11,0x11,0x11,0x11,0x11,0x11,0x11	;21.-40. Pixel, rot
.db 0x22,0x22,0x22,0x22,0x22,0x22,0x22,0x22,0x22,0x22	;41.-60. Pixel, gruen
.db 0x44,0x44,0x44,0x44,0x44,0x44,0x44,0x44,0x44,0x44	;61.-80. Pixel, blau
.db 0x77,0x77,0x77,0x77,0x77,0x77,0x77,0x77,0x77,0x77	;21.-100. Pixel, weiss
.db 0x33,0x33,0x33,0x33,0x33,0x33,0x33,0x33,0x33,0x33	;21.-120. Pixel, gelb
.db 0x55,0x55,0x55,0x55,0x55,0x55,0x55,0x55,0x55,0x55	;21.-140. Pixel, cyan
.db 0x66,0x66,0x66,0x66,0x66,0x66,0x66,0x66,0x66,0x66	;21.-160. Pixel, rotblau
.db 0x70,0x70,0x70,0x70,0x70,0x70,0x70,0x70,0x70,0x70	;21.-180. Pixel, grau
.db  0x00,0x00,0x00,0x00,0x00,0x0F			;181.-192. Pixel, schwarz
;; 16. Zeile:					
.db 0xF0,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00	;erste 10 Bytes = 20 Pixel, schwarz
.db 0x11,0x11,0x11,0x11,0x11,0x11,0x11,0x11,0x11,0x11	;21.-40. Pixel, rot
.db 0x22,0x22,0x22,0x22,0x22,0x22,0x22,0x22,0x22,0x22	;41.-60. Pixel, gruen
.db 0x44,0x44,0x44,0x44,0x44,0x44,0x44,0x44,0x44,0x44	;61.-80. Pixel, blau
.db 0x77,0x77,0x77,0x77,0x77,0x77,0x77,0x77,0x77,0x77	;21.-100. Pixel, weiss
.db 0x33,0x33,0x33,0x33,0x33,0x33,0x33,0x33,0x33,0x33	;21.-120. Pixel, gelb
.db 0x55,0x55,0x55,0x55,0x55,0x55,0x55,0x55,0x55,0x55	;21.-140. Pixel, cyan
.db 0x66,0x66,0x66,0x66,0x66,0x66,0x66,0x66,0x66,0x66	;21.-160. Pixel, rotblau
.db 0x70,0x70,0x70,0x70,0x70,0x70,0x70,0x70,0x70,0x70	;21.-180. Pixel, grau
.db  0x00,0x00,0x00,0x00,0x00,0x0F			;181.-192. Pixel, schwarz
;; 17. Zeile:					
.db 0xF0,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00	;erste 10 Bytes = 20 Pixel, schwarz
.db 0x11,0x11,0x11,0x11,0x11,0x11,0x11,0x11,0x11,0x11	;21.-40. Pixel, rot
.db 0x22,0x22,0x22,0x22,0x22,0x22,0x22,0x22,0x22,0x22	;41.-60. Pixel, gruen
.db 0x44,0x44,0x44,0x44,0x44,0x44,0x44,0x44,0x44,0x44	;61.-80. Pixel, blau
.db 0x77,0x77,0x77,0x77,0x77,0x77,0x77,0x77,0x77,0x77	;21.-100. Pixel, weiss
.db 0x33,0x33,0x33,0x33,0x33,0x33,0x33,0x33,0x33,0x33	;21.-120. Pixel, gelb
.db 0x55,0x55,0x55,0x55,0x55,0x55,0x55,0x55,0x55,0x55	;21.-140. Pixel, cyan
.db 0x66,0x66,0x66,0x66,0x66,0x66,0x66,0x66,0x66,0x66	;21.-160. Pixel, rotblau
.db 0x70,0x70,0x70,0x70,0x70,0x70,0x70,0x70,0x70,0x70	;21.-180. Pixel, grau
.db  0x00,0x00,0x00,0x00,0x00,0x0F			;181.-192. Pixel, schwarz
;; 18. Zeile:					
.db 0xF0,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00	;erste 10 Bytes = 20 Pixel, schwarz
.db 0x11,0x11,0x11,0x11,0x11,0x11,0x11,0x11,0x11,0x11	;21.-40. Pixel, rot
.db 0x22,0x22,0x22,0x22,0x22,0x22,0x22,0x22,0x22,0x22	;41.-60. Pixel, gruen
.db 0x44,0x44,0x44,0x44,0x44,0x44,0x44,0x44,0x44,0x44	;61.-80. Pixel, blau
.db 0x77,0x77,0x77,0x77,0x77,0x77,0x77,0x77,0x77,0x77	;21.-100. Pixel, weiss
.db 0x33,0x33,0x33,0x33,0x33,0x33,0x33,0x33,0x33,0x33	;21.-120. Pixel, gelb
.db 0x55,0x55,0x55,0x55,0x55,0x55,0x55,0x55,0x55,0x55	;21.-140. Pixel, cyan
.db 0x66,0x66,0x66,0x66,0x66,0x66,0x66,0x66,0x66,0x66	;21.-160. Pixel, rotblau
.db 0x70,0x70,0x70,0x70,0x70,0x70,0x70,0x70,0x70,0x70	;21.-180. Pixel, grau
.db  0x00,0x00,0x00,0x00,0x00,0x0F			;181.-192. Pixel, schwarz
;; 19. Zeile:					
.db 0xF0,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00	;erste 10 Bytes = 20 Pixel, schwarz
.db 0x11,0x11,0x11,0x11,0x11,0x11,0x11,0x11,0x11,0x11	;21.-40. Pixel, rot
.db 0x22,0x22,0x22,0x22,0x22,0x22,0x22,0x22,0x22,0x22	;41.-60. Pixel, gruen
.db 0x44,0x44,0x44,0x44,0x44,0x44,0x44,0x44,0x44,0x44	;61.-80. Pixel, blau
.db 0x77,0x77,0x77,0x77,0x77,0x77,0x77,0x77,0x77,0x77	;21.-100. Pixel, weiss
.db 0x33,0x33,0x33,0x33,0x33,0x33,0x33,0x33,0x33,0x33	;21.-120. Pixel, gelb
.db 0x55,0x55,0x55,0x55,0x55,0x55,0x55,0x55,0x55,0x55	;21.-140. Pixel, cyan
.db 0x66,0x66,0x66,0x66,0x66,0x66,0x66,0x66,0x66,0x66	;21.-160. Pixel, rotblau
.db 0x70,0x70,0x70,0x70,0x70,0x70,0x70,0x70,0x70,0x70	;21.-180. Pixel, grau
.db  0x00,0x00,0x00,0x00,0x00,0x0F			;181.-192. Pixel, schwarz
;; 20. Zeile:					
.db 0xF0,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00	;erste 10 Bytes = 20 Pixel, schwarz
.db 0x11,0x11,0x11,0x11,0x11,0x11,0x11,0x11,0x11,0x11	;21.-40. Pixel, rot
.db 0x22,0x22,0x22,0x22,0x22,0x22,0x22,0x22,0x22,0x22	;41.-60. Pixel, gruen
.db 0x44,0x44,0x44,0x44,0x44,0x44,0x44,0x44,0x44,0x44	;61.-80. Pixel, blau
.db 0x77,0x77,0x77,0x77,0x77,0x77,0x77,0x77,0x77,0x77	;21.-100. Pixel, weiss
.db 0x33,0x33,0x33,0x33,0x33,0x33,0x33,0x33,0x33,0x33	;21.-120. Pixel, gelb
.db 0x55,0x55,0x55,0x55,0x55,0x55,0x55,0x55,0x55,0x55	;21.-140. Pixel, cyan
.db 0x66,0x66,0x66,0x66,0x66,0x66,0x66,0x66,0x66,0x66	;21.-160. Pixel, rotblau
.db 0x70,0x70,0x70,0x70,0x70,0x70,0x70,0x70,0x70,0x70	;21.-180. Pixel, grau
.db  0x00,0x00,0x00,0x00,0x00,0x0F			;181.-192. Pixel, schwarz
;; 21.-40. Zeile:
.db 0xF0,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00	;erste 10 Bytes = 20 Pixel, schwarz
.db 0x11,0x11,0x11,0x11,0x11,0x11,0x11,0x11,0x11,0x11	;21.-40. Pixel, rot
.db 0x22,0x22,0x22,0x22,0x22,0x22,0x22,0x22,0x22,0x22	;41.-60. Pixel, gruen
.db 0x44,0x44,0x44,0x44,0x44,0x44,0x44,0x44,0x44,0x44	;61.-80. Pixel, blau
.db 0x77,0x77,0x77,0x77,0x77,0x77,0x77,0x77,0x77,0x77	;21.-100. Pixel, weiss
.db 0x33,0x33,0x33,0x33,0x33,0x33,0x33,0x33,0x33,0x33	;21.-120. Pixel, gelb
.db 0x55,0x55,0x55,0x55,0x55,0x55,0x55,0x55,0x55,0x55	;21.-140. Pixel, cyan
.db 0x66,0x66,0x66,0x66,0x66,0x66,0x66,0x66,0x66,0x66	;21.-160. Pixel, rotblau
.db 0x70,0x70,0x70,0x70,0x70,0x70,0x70,0x70,0x70,0x70	;21.-180. Pixel, grau
.db  0x00,0x00,0x00,0x00,0x00,0x0F			;181.-192. Pixel, schwarz
;; 2. Zeile:					
.db 0xF0,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00	;erste 10 Bytes = 20 Pixel, schwarz
.db 0x11,0x11,0x11,0x11,0x11,0x11,0x11,0x11,0x11,0x11	;21.-40. Pixel, rot
.db 0x22,0x22,0x22,0x22,0x22,0x22,0x22,0x22,0x22,0x22	;41.-60. Pixel, gruen
.db 0x44,0x44,0x44,0x44,0x44,0x44,0x44,0x44,0x44,0x44	;61.-80. Pixel, blau
.db 0x77,0x77,0x77,0x77,0x77,0x77,0x77,0x77,0x77,0x77	;21.-100. Pixel, weiss
.db 0x33,0x33,0x33,0x33,0x33,0x33,0x33,0x33,0x33,0x33	;21.-120. Pixel, gelb
.db 0x55,0x55,0x55,0x55,0x55,0x55,0x55,0x55,0x55,0x55	;21.-140. Pixel, cyan
.db 0x66,0x66,0x66,0x66,0x66,0x66,0x66,0x66,0x66,0x66	;21.-160. Pixel, rotblau
.db 0x70,0x70,0x70,0x70,0x70,0x70,0x70,0x70,0x70,0x70	;21.-180. Pixel, grau
.db  0x00,0x00,0x00,0x00,0x00,0x0F			;181.-192. Pixel, schwarz
;; 3. Zeile:					
.db 0xF0,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00	;erste 10 Bytes = 20 Pixel, schwarz
.db 0x11,0x11,0x11,0x11,0x11,0x11,0x11,0x11,0x11,0x11	;21.-40. Pixel, rot
.db 0x22,0x22,0x22,0x22,0x22,0x22,0x22,0x22,0x22,0x22	;41.-60. Pixel, gruen
.db 0x44,0x44,0x44,0x44,0x44,0x44,0x44,0x44,0x44,0x44	;61.-80. Pixel, blau
.db 0x77,0x77,0x77,0x77,0x77,0x77,0x77,0x77,0x77,0x77	;21.-100. Pixel, weiss
.db 0x33,0x33,0x33,0x33,0x33,0x33,0x33,0x33,0x33,0x33	;21.-120. Pixel, gelb
.db 0x55,0x55,0x55,0x55,0x55,0x55,0x55,0x55,0x55,0x55	;21.-140. Pixel, cyan
.db 0x66,0x66,0x66,0x66,0x66,0x66,0x66,0x66,0x66,0x66	;21.-160. Pixel, rotblau
.db 0x70,0x70,0x70,0x70,0x70,0x70,0x70,0x70,0x70,0x70	;21.-180. Pixel, grau
.db  0x00,0x00,0x00,0x00,0x00,0x0F			;181.-192. Pixel, schwarz
;; 4. Zeile:					
.db 0xF0,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00	;erste 10 Bytes = 20 Pixel, schwarz
.db 0x11,0x11,0x11,0x11,0x11,0x11,0x11,0x11,0x11,0x11	;21.-40. Pixel, rot
.db 0x22,0x22,0x22,0x22,0x22,0x22,0x22,0x22,0x22,0x22	;41.-60. Pixel, gruen
.db 0x44,0x44,0x44,0x44,0x44,0x44,0x44,0x44,0x44,0x44	;61.-80. Pixel, blau
.db 0x77,0x77,0x77,0x77,0x77,0x77,0x77,0x77,0x77,0x77	;21.-100. Pixel, weiss
.db 0x33,0x33,0x33,0x33,0x33,0x33,0x33,0x33,0x33,0x33	;21.-120. Pixel, gelb
.db 0x55,0x55,0x55,0x55,0x55,0x55,0x55,0x55,0x55,0x55	;21.-140. Pixel, cyan
.db 0x66,0x66,0x66,0x66,0x66,0x66,0x66,0x66,0x66,0x66	;21.-160. Pixel, rotblau
.db 0x70,0x70,0x70,0x70,0x70,0x70,0x70,0x70,0x70,0x70	;21.-180. Pixel, grau
.db  0x00,0x00,0x00,0x00,0x00,0x0F			;181.-192. Pixel, schwarz
;; 5. Zeile:					
.db 0xF0,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00	;erste 10 Bytes = 20 Pixel, schwarz
.db 0x11,0x11,0x11,0x11,0x11,0x11,0x11,0x11,0x11,0x11	;21.-40. Pixel, rot
.db 0x22,0x22,0x22,0x22,0x22,0x22,0x22,0x22,0x22,0x22	;41.-60. Pixel, gruen
.db 0x44,0x44,0x44,0x44,0x44,0x44,0x44,0x44,0x44,0x44	;61.-80. Pixel, blau
.db 0x77,0x77,0x77,0x77,0x77,0x77,0x77,0x77,0x77,0x77	;21.-100. Pixel, weiss
.db 0x33,0x33,0x33,0x33,0x33,0x33,0x33,0x33,0x33,0x33	;21.-120. Pixel, gelb
.db 0x55,0x55,0x55,0x55,0x55,0x55,0x55,0x55,0x55,0x55	;21.-140. Pixel, cyan
.db 0x66,0x66,0x66,0x66,0x66,0x66,0x66,0x66,0x66,0x66	;21.-160. Pixel, rotblau
.db 0x70,0x70,0x70,0x70,0x70,0x70,0x70,0x70,0x70,0x70	;21.-180. Pixel, grau
.db  0x00,0x00,0x00,0x00,0x00,0x0F			;181.-192. Pixel, schwarz
;; 6. Zeile:					
.db 0xF0,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00	;erste 10 Bytes = 20 Pixel, schwarz
.db 0x11,0x11,0x11,0x11,0x11,0x11,0x11,0x11,0x11,0x11	;21.-40. Pixel, rot
.db 0x22,0x22,0x22,0x22,0x22,0x22,0x22,0x22,0x22,0x22	;41.-60. Pixel, gruen
.db 0x44,0x44,0x44,0x44,0x44,0x44,0x44,0x44,0x44,0x44	;61.-80. Pixel, blau
.db 0x77,0x77,0x77,0x77,0x77,0x77,0x77,0x77,0x77,0x77	;21.-100. Pixel, weiss
.db 0x33,0x33,0x33,0x33,0x33,0x33,0x33,0x33,0x33,0x33	;21.-120. Pixel, gelb
.db 0x55,0x55,0x55,0x55,0x55,0x55,0x55,0x55,0x55,0x55	;21.-140. Pixel, cyan
.db 0x66,0x66,0x66,0x66,0x66,0x66,0x66,0x66,0x66,0x66	;21.-160. Pixel, rotblau
.db 0x70,0x70,0x70,0x70,0x70,0x70,0x70,0x70,0x70,0x70	;21.-180. Pixel, grau
.db  0x00,0x00,0x00,0x00,0x00,0x0F			;181.-192. Pixel, schwarz
;; 7. Zeile:					
.db 0xF0,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00	;erste 10 Bytes = 20 Pixel, schwarz
.db 0x11,0x11,0x11,0x11,0x11,0x11,0x11,0x11,0x11,0x11	;21.-40. Pixel, rot
.db 0x22,0x22,0x22,0x22,0x22,0x22,0x22,0x22,0x22,0x22	;41.-60. Pixel, gruen
.db 0x44,0x44,0x44,0x44,0x44,0x44,0x44,0x44,0x44,0x44	;61.-80. Pixel, blau
.db 0x77,0x77,0x77,0x77,0x77,0x77,0x77,0x77,0x77,0x77	;21.-100. Pixel, weiss
.db 0x33,0x33,0x33,0x33,0x33,0x33,0x33,0x33,0x33,0x33	;21.-120. Pixel, gelb
.db 0x55,0x55,0x55,0x55,0x55,0x55,0x55,0x55,0x55,0x55	;21.-140. Pixel, cyan
.db 0x66,0x66,0x66,0x66,0x66,0x66,0x66,0x66,0x66,0x66	;21.-160. Pixel, rotblau
.db 0x70,0x70,0x70,0x70,0x70,0x70,0x70,0x70,0x70,0x70	;21.-180. Pixel, grau
.db  0x00,0x00,0x00,0x00,0x00,0x0F			;181.-192. Pixel, schwarz
;; 8. Zeile:					
.db 0xF0,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00	;erste 10 Bytes = 20 Pixel, schwarz
.db 0x11,0x11,0x11,0x11,0x11,0x11,0x11,0x11,0x11,0x11	;21.-40. Pixel, rot
.db 0x22,0x22,0x22,0x22,0x22,0x22,0x22,0x22,0x22,0x22	;41.-60. Pixel, gruen
.db 0x44,0x44,0x44,0x44,0x44,0x44,0x44,0x44,0x44,0x44	;61.-80. Pixel, blau
.db 0x77,0x77,0x77,0x77,0x77,0x77,0x77,0x77,0x77,0x77	;21.-100. Pixel, weiss
.db 0x33,0x33,0x33,0x33,0x33,0x33,0x33,0x33,0x33,0x33	;21.-120. Pixel, gelb
.db 0x55,0x55,0x55,0x55,0x55,0x55,0x55,0x55,0x55,0x55	;21.-140. Pixel, cyan
.db 0x66,0x66,0x66,0x66,0x66,0x66,0x66,0x66,0x66,0x66	;21.-160. Pixel, rotblau
.db 0x70,0x70,0x70,0x70,0x70,0x70,0x70,0x70,0x70,0x70	;21.-180. Pixel, grau
.db  0x00,0x00,0x00,0x00,0x00,0x0F			;181.-192. Pixel, schwarz
;; 9. Zeile:					
.db 0xF0,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00	;erste 10 Bytes = 20 Pixel, schwarz
.db 0x11,0x11,0x11,0x11,0x11,0x11,0x11,0x11,0x11,0x11	;21.-40. Pixel, rot
.db 0x22,0x22,0x22,0x22,0x22,0x22,0x22,0x22,0x22,0x22	;41.-60. Pixel, gruen
.db 0x44,0x44,0x44,0x44,0x44,0x44,0x44,0x44,0x44,0x44	;61.-80. Pixel, blau
.db 0x77,0x77,0x77,0x77,0x77,0x77,0x77,0x77,0x77,0x77	;21.-100. Pixel, weiss
.db 0x33,0x33,0x33,0x33,0x33,0x33,0x33,0x33,0x33,0x33	;21.-120. Pixel, gelb
.db 0x55,0x55,0x55,0x55,0x55,0x55,0x55,0x55,0x55,0x55	;21.-140. Pixel, cyan
.db 0x66,0x66,0x66,0x66,0x66,0x66,0x66,0x66,0x66,0x66	;21.-160. Pixel, rotblau
.db 0x70,0x70,0x70,0x70,0x70,0x70,0x70,0x70,0x70,0x70	;21.-180. Pixel, grau
.db  0x00,0x00,0x00,0x00,0x00,0x0F			;181.-192. Pixel, schwarz
;; 10. Zeile:					
.db 0xF0,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00	;erste 10 Bytes = 20 Pixel, schwarz
.db 0x11,0x11,0x11,0x11,0x11,0x11,0x11,0x11,0x11,0x11	;21.-40. Pixel, rot
.db 0x22,0x22,0x22,0x22,0x22,0x22,0x22,0x22,0x22,0x22	;41.-60. Pixel, gruen
.db 0x44,0x44,0x44,0x44,0x44,0x44,0x44,0x44,0x44,0x44	;61.-80. Pixel, blau
.db 0x77,0x77,0x77,0x77,0x77,0x77,0x77,0x77,0x77,0x77	;21.-100. Pixel, weiss
.db 0x33,0x33,0x33,0x33,0x33,0x33,0x33,0x33,0x33,0x33	;21.-120. Pixel, gelb
.db 0x55,0x55,0x55,0x55,0x55,0x55,0x55,0x55,0x55,0x55	;21.-140. Pixel, cyan
.db 0x66,0x66,0x66,0x66,0x66,0x66,0x66,0x66,0x66,0x66	;21.-160. Pixel, rotblau
.db 0x70,0x70,0x70,0x70,0x70,0x70,0x70,0x70,0x70,0x70	;21.-180. Pixel, grau
.db  0x00,0x00,0x00,0x00,0x00,0x0F			;181.-192. Pixel, schwarz
;; 11.-20. Zeile:
.db 0xF0,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00	;erste 10 Bytes = 20 Pixel, schwarz
.db 0x11,0x11,0x11,0x11,0x11,0x11,0x11,0x11,0x11,0x11	;21.-40. Pixel, rot
.db 0x22,0x22,0x22,0x22,0x22,0x22,0x22,0x22,0x22,0x22	;41.-60. Pixel, gruen
.db 0x44,0x44,0x44,0x44,0x44,0x44,0x44,0x44,0x44,0x44	;61.-80. Pixel, blau
.db 0x77,0x77,0x77,0x77,0x77,0x77,0x77,0x77,0x77,0x77	;21.-100. Pixel, weiss
.db 0x33,0x33,0x33,0x33,0x33,0x33,0x33,0x33,0x33,0x33	;21.-120. Pixel, gelb
.db 0x55,0x55,0x55,0x55,0x55,0x55,0x55,0x55,0x55,0x55	;21.-140. Pixel, cyan
.db 0x66,0x66,0x66,0x66,0x66,0x66,0x66,0x66,0x66,0x66	;21.-160. Pixel, rotblau
.db 0x70,0x70,0x70,0x70,0x70,0x70,0x70,0x70,0x70,0x70	;21.-180. Pixel, grau
.db  0x00,0x00,0x00,0x00,0x00,0x0F			;181.-192. Pixel, schwarz
;; 12. Zeile:					
.db 0xF0,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00	;erste 10 Bytes = 20 Pixel, schwarz
.db 0x11,0x11,0x11,0x11,0x11,0x11,0x11,0x11,0x11,0x11	;21.-40. Pixel, rot
.db 0x22,0x22,0x22,0x22,0x22,0x22,0x22,0x22,0x22,0x22	;41.-60. Pixel, gruen
.db 0x44,0x44,0x44,0x44,0x44,0x44,0x44,0x44,0x44,0x44	;61.-80. Pixel, blau
.db 0x77,0x77,0x77,0x77,0x77,0x77,0x77,0x77,0x77,0x77	;21.-100. Pixel, weiss
.db 0x33,0x33,0x33,0x33,0x33,0x33,0x33,0x33,0x33,0x33	;21.-120. Pixel, gelb
.db 0x55,0x55,0x55,0x55,0x55,0x55,0x55,0x55,0x55,0x55	;21.-140. Pixel, cyan
.db 0x66,0x66,0x66,0x66,0x66,0x66,0x66,0x66,0x66,0x66	;21.-160. Pixel, rotblau
.db 0x70,0x70,0x70,0x70,0x70,0x70,0x70,0x70,0x70,0x70	;21.-180. Pixel, grau
.db  0x00,0x00,0x00,0x00,0x00,0x0F			;181.-192. Pixel, schwarz
;; 13. Zeile:					
.db 0xF0,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00	;erste 10 Bytes = 20 Pixel, schwarz
.db 0x11,0x11,0x11,0x11,0x11,0x11,0x11,0x11,0x11,0x11	;21.-40. Pixel, rot
.db 0x22,0x22,0x22,0x22,0x22,0x22,0x22,0x22,0x22,0x22	;41.-60. Pixel, gruen
.db 0x44,0x44,0x44,0x44,0x44,0x44,0x44,0x44,0x44,0x44	;61.-80. Pixel, blau
.db 0x77,0x77,0x77,0x77,0x77,0x77,0x77,0x77,0x77,0x77	;21.-100. Pixel, weiss
.db 0x33,0x33,0x33,0x33,0x33,0x33,0x33,0x33,0x33,0x33	;21.-120. Pixel, gelb
.db 0x55,0x55,0x55,0x55,0x55,0x55,0x55,0x55,0x55,0x55	;21.-140. Pixel, cyan
.db 0x66,0x66,0x66,0x66,0x66,0x66,0x66,0x66,0x66,0x66	;21.-160. Pixel, rotblau
.db 0x70,0x70,0x70,0x70,0x70,0x70,0x70,0x70,0x70,0x70	;21.-180. Pixel, grau
.db  0x00,0x00,0x00,0x00,0x00,0x0F			;181.-192. Pixel, schwarz
;; 14. Zeile:					
.db 0xF0,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00	;erste 10 Bytes = 20 Pixel, schwarz
.db 0x11,0x11,0x11,0x11,0x11,0x11,0x11,0x11,0x11,0x11	;21.-40. Pixel, rot
.db 0x22,0x22,0x22,0x22,0x22,0x22,0x22,0x22,0x22,0x22	;41.-60. Pixel, gruen
.db 0x44,0x44,0x44,0x44,0x44,0x44,0x44,0x44,0x44,0x44	;61.-80. Pixel, blau
.db 0x77,0x77,0x77,0x77,0x77,0x77,0x77,0x77,0x77,0x77	;21.-100. Pixel, weiss
.db 0x33,0x33,0x33,0x33,0x33,0x33,0x33,0x33,0x33,0x33	;21.-120. Pixel, gelb
.db 0x55,0x55,0x55,0x55,0x55,0x55,0x55,0x55,0x55,0x55	;21.-140. Pixel, cyan
.db 0x66,0x66,0x66,0x66,0x66,0x66,0x66,0x66,0x66,0x66	;21.-160. Pixel, rotblau
.db 0x70,0x70,0x70,0x70,0x70,0x70,0x70,0x70,0x70,0x70	;21.-180. Pixel, grau
.db  0x00,0x00,0x00,0x00,0x00,0x0F			;181.-192. Pixel, schwarz
;; 15. Zeile:					
.db 0xF0,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00	;erste 10 Bytes = 20 Pixel, schwarz
.db 0x11,0x11,0x11,0x11,0x11,0x11,0x11,0x11,0x11,0x11	;21.-40. Pixel, rot
.db 0x22,0x22,0x22,0x22,0x22,0x22,0x22,0x22,0x22,0x22	;41.-60. Pixel, gruen
.db 0x44,0x44,0x44,0x44,0x44,0x44,0x44,0x44,0x44,0x44	;61.-80. Pixel, blau
.db 0x77,0x77,0x77,0x77,0x77,0x77,0x77,0x77,0x77,0x77	;21.-100. Pixel, weiss
.db 0x33,0x33,0x33,0x33,0x33,0x33,0x33,0x33,0x33,0x33	;21.-120. Pixel, gelb
.db 0x55,0x55,0x55,0x55,0x55,0x55,0x55,0x55,0x55,0x55	;21.-140. Pixel, cyan
.db 0x66,0x66,0x66,0x66,0x66,0x66,0x66,0x66,0x66,0x66	;21.-160. Pixel, rotblau
.db 0x70,0x70,0x70,0x70,0x70,0x70,0x70,0x70,0x70,0x70	;21.-180. Pixel, grau
.db  0x00,0x00,0x00,0x00,0x00,0x0F			;181.-192. Pixel, schwarz
;; 16. Zeile:					
.db 0xF0,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00	;erste 10 Bytes = 20 Pixel, schwarz
.db 0x11,0x11,0x11,0x11,0x11,0x11,0x11,0x11,0x11,0x11	;21.-40. Pixel, rot
.db 0x22,0x22,0x22,0x22,0x22,0x22,0x22,0x22,0x22,0x22	;41.-60. Pixel, gruen
.db 0x44,0x44,0x44,0x44,0x44,0x44,0x44,0x44,0x44,0x44	;61.-80. Pixel, blau
.db 0x77,0x77,0x77,0x77,0x77,0x77,0x77,0x77,0x77,0x77	;21.-100. Pixel, weiss
.db 0x33,0x33,0x33,0x33,0x33,0x33,0x33,0x33,0x33,0x33	;21.-120. Pixel, gelb
.db 0x55,0x55,0x55,0x55,0x55,0x55,0x55,0x55,0x55,0x55	;21.-140. Pixel, cyan
.db 0x66,0x66,0x66,0x66,0x66,0x66,0x66,0x66,0x66,0x66	;21.-160. Pixel, rotblau
.db 0x70,0x70,0x70,0x70,0x70,0x70,0x70,0x70,0x70,0x70	;21.-180. Pixel, grau
.db  0x00,0x00,0x00,0x00,0x00,0x0F			;181.-192. Pixel, schwarz
;; 17. Zeile:					
.db 0xF0,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00	;erste 10 Bytes = 20 Pixel, schwarz
.db 0x11,0x11,0x11,0x11,0x11,0x11,0x11,0x11,0x11,0x11	;21.-40. Pixel, rot
.db 0x22,0x22,0x22,0x22,0x22,0x22,0x22,0x22,0x22,0x22	;41.-60. Pixel, gruen
.db 0x44,0x44,0x44,0x44,0x44,0x44,0x44,0x44,0x44,0x44	;61.-80. Pixel, blau
.db 0x77,0x77,0x77,0x77,0x77,0x77,0x77,0x77,0x77,0x77	;21.-100. Pixel, weiss
.db 0x33,0x33,0x33,0x33,0x33,0x33,0x33,0x33,0x33,0x33	;21.-120. Pixel, gelb
.db 0x55,0x55,0x55,0x55,0x55,0x55,0x55,0x55,0x55,0x55	;21.-140. Pixel, cyan
.db 0x66,0x66,0x66,0x66,0x66,0x66,0x66,0x66,0x66,0x66	;21.-160. Pixel, rotblau
.db 0x70,0x70,0x70,0x70,0x70,0x70,0x70,0x70,0x70,0x70	;21.-180. Pixel, grau
.db  0x00,0x00,0x00,0x00,0x00,0x0F			;181.-192. Pixel, schwarz
;; 18. Zeile:					
.db 0xF0,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00	;erste 10 Bytes = 20 Pixel, schwarz
.db 0x11,0x11,0x11,0x11,0x11,0x11,0x11,0x11,0x11,0x11	;21.-40. Pixel, rot
.db 0x22,0x22,0x22,0x22,0x22,0x22,0x22,0x22,0x22,0x22	;41.-60. Pixel, gruen
.db 0x44,0x44,0x44,0x44,0x44,0x44,0x44,0x44,0x44,0x44	;61.-80. Pixel, blau
.db 0x77,0x77,0x77,0x77,0x77,0x77,0x77,0x77,0x77,0x77	;21.-100. Pixel, weiss
.db 0x33,0x33,0x33,0x33,0x33,0x33,0x33,0x33,0x33,0x33	;21.-120. Pixel, gelb
.db 0x55,0x55,0x55,0x55,0x55,0x55,0x55,0x55,0x55,0x55	;21.-140. Pixel, cyan
.db 0x66,0x66,0x66,0x66,0x66,0x66,0x66,0x66,0x66,0x66	;21.-160. Pixel, rotblau
.db 0x70,0x70,0x70,0x70,0x70,0x70,0x70,0x70,0x70,0x70	;21.-180. Pixel, grau
.db  0x00,0x00,0x00,0x00,0x00,0x0F			;181.-192. Pixel, schwarz
;; 19. Zeile:					
.db 0xF0,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00	;erste 10 Bytes = 20 Pixel, schwarz
.db 0x11,0x11,0x11,0x11,0x11,0x11,0x11,0x11,0x11,0x11	;21.-40. Pixel, rot
.db 0x22,0x22,0x22,0x22,0x22,0x22,0x22,0x22,0x22,0x22	;41.-60. Pixel, gruen
.db 0x44,0x44,0x44,0x44,0x44,0x44,0x44,0x44,0x44,0x44	;61.-80. Pixel, blau
.db 0x77,0x77,0x77,0x77,0x77,0x77,0x77,0x77,0x77,0x77	;21.-100. Pixel, weiss
.db 0x33,0x33,0x33,0x33,0x33,0x33,0x33,0x33,0x33,0x33	;21.-120. Pixel, gelb
.db 0x55,0x55,0x55,0x55,0x55,0x55,0x55,0x55,0x55,0x55	;21.-140. Pixel, cyan
.db 0x66,0x66,0x66,0x66,0x66,0x66,0x66,0x66,0x66,0x66	;21.-160. Pixel, rotblau
.db 0x70,0x70,0x70,0x70,0x70,0x70,0x70,0x70,0x70,0x70	;21.-180. Pixel, grau
.db  0x00,0x00,0x00,0x00,0x00,0x0F			;181.-192. Pixel, schwarz
;; 40. Zeile:
.db 0xF0,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00	;erste 10 Bytes = 20 Pixel, schwarz
.db 0x11,0x11,0x11,0x11,0x11,0x11,0x11,0x11,0x11,0x11	;21.-40. Pixel, rot
.db 0x22,0x22,0x22,0x22,0x22,0x22,0x22,0x22,0x22,0x22	;41.-60. Pixel, gruen
.db 0x44,0x44,0x44,0x44,0x44,0x44,0x44,0x44,0x44,0x44	;61.-80. Pixel, blau
.db 0x77,0x77,0x77,0x77,0x77,0x77,0x77,0x77,0x77,0x77	;21.-100. Pixel, weiss
.db 0x33,0x33,0x33,0x33,0x33,0x33,0x33,0x33,0x33,0x33	;21.-120. Pixel, gelb
.db 0x55,0x55,0x55,0x55,0x55,0x55,0x55,0x55,0x55,0x55	;21.-140. Pixel, cyan
.db 0x66,0x66,0x66,0x66,0x66,0x66,0x66,0x66,0x66,0x66	;21.-160. Pixel, rotblau
.db 0x70,0x70,0x70,0x70,0x70,0x70,0x70,0x70,0x70,0x70	;21.-180. Pixel, grau
.db  0x00,0x00,0x00,0x00,0x00,0x0F			;181.-192. Pixel, schwarz
;; 21.-40. Zeile:
.db 0xF0,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00	;erste 10 Bytes = 20 Pixel, schwarz
.db 0x11,0x11,0x11,0x11,0x11,0x11,0x11,0x11,0x11,0x11	;21.-40. Pixel, rot
.db 0x22,0x22,0x22,0x22,0x22,0x22,0x22,0x22,0x22,0x22	;41.-60. Pixel, gruen
.db 0x44,0x44,0x44,0x44,0x44,0x44,0x44,0x44,0x44,0x44	;61.-80. Pixel, blau
.db 0x77,0x77,0x77,0x77,0x77,0x77,0x77,0x77,0x77,0x77	;21.-100. Pixel, weiss
.db 0x33,0x33,0x33,0x33,0x33,0x33,0x33,0x33,0x33,0x33	;21.-120. Pixel, gelb
.db 0x55,0x55,0x55,0x55,0x55,0x55,0x55,0x55,0x55,0x55	;21.-140. Pixel, cyan
.db 0x66,0x66,0x66,0x66,0x66,0x66,0x66,0x66,0x66,0x66	;21.-160. Pixel, rotblau
.db 0x70,0x70,0x70,0x70,0x70,0x70,0x70,0x70,0x70,0x70	;21.-180. Pixel, grau
.db  0x00,0x00,0x00,0x00,0x00,0x0F			;181.-192. Pixel, schwarz
;; 2. Zeile:					
.db 0xF0,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00	;erste 10 Bytes = 20 Pixel, schwarz
.db 0x11,0x11,0x11,0x11,0x11,0x11,0x11,0x11,0x11,0x11	;21.-40. Pixel, rot
.db 0x22,0x22,0x22,0x22,0x22,0x22,0x22,0x22,0x22,0x22	;41.-60. Pixel, gruen
.db 0x44,0x44,0x44,0x44,0x44,0x44,0x44,0x44,0x44,0x44	;61.-80. Pixel, blau
.db 0x77,0x77,0x77,0x77,0x77,0x77,0x77,0x77,0x77,0x77	;21.-100. Pixel, weiss
.db 0x33,0x33,0x33,0x33,0x33,0x33,0x33,0x33,0x33,0x33	;21.-120. Pixel, gelb
.db 0x55,0x55,0x55,0x55,0x55,0x55,0x55,0x55,0x55,0x55	;21.-140. Pixel, cyan
.db 0x66,0x66,0x66,0x66,0x66,0x66,0x66,0x66,0x66,0x66	;21.-160. Pixel, rotblau
.db 0x70,0x70,0x70,0x70,0x70,0x70,0x70,0x70,0x70,0x70	;21.-180. Pixel, grau
.db  0x00,0x00,0x00,0x00,0x00,0x0F			;181.-192. Pixel, schwarz
;; 3. Zeile:					
.db 0xF0,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00	;erste 10 Bytes = 20 Pixel, schwarz
.db 0x11,0x11,0x11,0x11,0x11,0x11,0x11,0x11,0x11,0x11	;21.-40. Pixel, rot
.db 0x22,0x22,0x22,0x22,0x22,0x22,0x22,0x22,0x22,0x22	;41.-60. Pixel, gruen
.db 0x44,0x44,0x44,0x44,0x44,0x44,0x44,0x44,0x44,0x44	;61.-80. Pixel, blau
.db 0x77,0x77,0x77,0x77,0x77,0x77,0x77,0x77,0x77,0x77	;21.-100. Pixel, weiss
.db 0x33,0x33,0x33,0x33,0x33,0x33,0x33,0x33,0x33,0x33	;21.-120. Pixel, gelb
.db 0x55,0x55,0x55,0x55,0x55,0x55,0x55,0x55,0x55,0x55	;21.-140. Pixel, cyan
.db 0x66,0x66,0x66,0x66,0x66,0x66,0x66,0x66,0x66,0x66	;21.-160. Pixel, rotblau
.db 0x70,0x70,0x70,0x70,0x70,0x70,0x70,0x70,0x70,0x70	;21.-180. Pixel, grau
.db  0x00,0x00,0x00,0x00,0x00,0x0F			;181.-192. Pixel, schwarz
;; 4. Zeile:					
.db 0xF0,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00	;erste 10 Bytes = 20 Pixel, schwarz
.db 0x11,0x11,0x11,0x11,0x11,0x11,0x11,0x11,0x11,0x11	;21.-40. Pixel, rot
.db 0x22,0x22,0x22,0x22,0x22,0x22,0x22,0x22,0x22,0x22	;41.-60. Pixel, gruen
.db 0x44,0x44,0x44,0x44,0x44,0x44,0x44,0x44,0x44,0x44	;61.-80. Pixel, blau
.db 0x77,0x77,0x77,0x77,0x77,0x77,0x77,0x77,0x77,0x77	;21.-100. Pixel, weiss
.db 0x33,0x33,0x33,0x33,0x33,0x33,0x33,0x33,0x33,0x33	;21.-120. Pixel, gelb
.db 0x55,0x55,0x55,0x55,0x55,0x55,0x55,0x55,0x55,0x55	;21.-140. Pixel, cyan
.db 0x66,0x66,0x66,0x66,0x66,0x66,0x66,0x66,0x66,0x66	;21.-160. Pixel, rotblau
.db 0x70,0x70,0x70,0x70,0x70,0x70,0x70,0x70,0x70,0x70	;21.-180. Pixel, grau
.db  0x00,0x00,0x00,0x00,0x00,0x0F			;181.-192. Pixel, schwarz
;; 5. Zeile:					
.db 0xF0,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00	;erste 10 Bytes = 20 Pixel, schwarz
.db 0x11,0x11,0x11,0x11,0x11,0x11,0x11,0x11,0x11,0x11	;21.-40. Pixel, rot
.db 0x22,0x22,0x22,0x22,0x22,0x22,0x22,0x22,0x22,0x22	;41.-60. Pixel, gruen
.db 0x44,0x44,0x44,0x44,0x44,0x44,0x44,0x44,0x44,0x44	;61.-80. Pixel, blau
.db 0x77,0x77,0x77,0x77,0x77,0x77,0x77,0x77,0x77,0x77	;21.-100. Pixel, weiss
.db 0x33,0x33,0x33,0x33,0x33,0x33,0x33,0x33,0x33,0x33	;21.-120. Pixel, gelb
.db 0x55,0x55,0x55,0x55,0x55,0x55,0x55,0x55,0x55,0x55	;21.-140. Pixel, cyan
.db 0x66,0x66,0x66,0x66,0x66,0x66,0x66,0x66,0x66,0x66	;21.-160. Pixel, rotblau
.db 0x70,0x70,0x70,0x70,0x70,0x70,0x70,0x70,0x70,0x70	;21.-180. Pixel, grau
.db  0x00,0x00,0x00,0x00,0x00,0x0F			;181.-192. Pixel, schwarz
;; 6. Zeile:					
.db 0xF0,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00	;erste 10 Bytes = 20 Pixel, schwarz
.db 0x11,0x11,0x11,0x11,0x11,0x11,0x11,0x11,0x11,0x11	;21.-40. Pixel, rot
.db 0x22,0x22,0x22,0x22,0x22,0x22,0x22,0x22,0x22,0x22	;41.-60. Pixel, gruen
.db 0x44,0x44,0x44,0x44,0x44,0x44,0x44,0x44,0x44,0x44	;61.-80. Pixel, blau
.db 0x77,0x77,0x77,0x77,0x77,0x77,0x77,0x77,0x77,0x77	;21.-100. Pixel, weiss
.db 0x33,0x33,0x33,0x33,0x33,0x33,0x33,0x33,0x33,0x33	;21.-120. Pixel, gelb
.db 0x55,0x55,0x55,0x55,0x55,0x55,0x55,0x55,0x55,0x55	;21.-140. Pixel, cyan
.db 0x66,0x66,0x66,0x66,0x66,0x66,0x66,0x66,0x66,0x66	;21.-160. Pixel, rotblau
.db 0x70,0x70,0x70,0x70,0x70,0x70,0x70,0x70,0x70,0x70	;21.-180. Pixel, grau
.db  0x00,0x00,0x00,0x00,0x00,0x0F			;181.-192. Pixel, schwarz
;; 7. Zeile:					
.db 0xF0,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00	;erste 10 Bytes = 20 Pixel, schwarz
.db 0x11,0x11,0x11,0x11,0x11,0x11,0x11,0x11,0x11,0x11	;21.-40. Pixel, rot
.db 0x22,0x22,0x22,0x22,0x22,0x22,0x22,0x22,0x22,0x22	;41.-60. Pixel, gruen
.db 0x44,0x44,0x44,0x44,0x44,0x44,0x44,0x44,0x44,0x44	;61.-80. Pixel, blau
.db 0x77,0x77,0x77,0x77,0x77,0x77,0x77,0x77,0x77,0x77	;21.-100. Pixel, weiss
.db 0x33,0x33,0x33,0x33,0x33,0x33,0x33,0x33,0x33,0x33	;21.-120. Pixel, gelb
.db 0x55,0x55,0x55,0x55,0x55,0x55,0x55,0x55,0x55,0x55	;21.-140. Pixel, cyan
.db 0x66,0x66,0x66,0x66,0x66,0x66,0x66,0x66,0x66,0x66	;21.-160. Pixel, rotblau
.db 0x70,0x70,0x70,0x70,0x70,0x70,0x70,0x70,0x70,0x70	;21.-180. Pixel, grau
.db  0x00,0x00,0x00,0x00,0x00,0x0F			;181.-192. Pixel, schwarz
;; 8. Zeile:					
.db 0xF0,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00	;erste 10 Bytes = 20 Pixel, schwarz
.db 0x11,0x11,0x11,0x11,0x11,0x11,0x11,0x11,0x11,0x11	;21.-40. Pixel, rot
.db 0x22,0x22,0x22,0x22,0x22,0x22,0x22,0x22,0x22,0x22	;41.-60. Pixel, gruen
.db 0x44,0x44,0x44,0x44,0x44,0x44,0x44,0x44,0x44,0x44	;61.-80. Pixel, blau
.db 0x77,0x77,0x77,0x77,0x77,0x77,0x77,0x77,0x77,0x77	;21.-100. Pixel, weiss
.db 0x33,0x33,0x33,0x33,0x33,0x33,0x33,0x33,0x33,0x33	;21.-120. Pixel, gelb
.db 0x55,0x55,0x55,0x55,0x55,0x55,0x55,0x55,0x55,0x55	;21.-140. Pixel, cyan
.db 0x66,0x66,0x66,0x66,0x66,0x66,0x66,0x66,0x66,0x66	;21.-160. Pixel, rotblau
.db 0x70,0x70,0x70,0x70,0x70,0x70,0x70,0x70,0x70,0x70	;21.-180. Pixel, grau
.db  0x00,0x00,0x00,0x00,0x00,0x0F			;181.-192. Pixel, schwarz
;; 9. Zeile:					
.db 0xF0,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00	;erste 10 Bytes = 20 Pixel, schwarz
.db 0x11,0x11,0x11,0x11,0x11,0x11,0x11,0x11,0x11,0x11	;21.-40. Pixel, rot
.db 0x22,0x22,0x22,0x22,0x22,0x22,0x22,0x22,0x22,0x22	;41.-60. Pixel, gruen
.db 0x44,0x44,0x44,0x44,0x44,0x44,0x44,0x44,0x44,0x44	;61.-80. Pixel, blau
.db 0x77,0x77,0x77,0x77,0x77,0x77,0x77,0x77,0x77,0x77	;21.-100. Pixel, weiss
.db 0x33,0x33,0x33,0x33,0x33,0x33,0x33,0x33,0x33,0x33	;21.-120. Pixel, gelb
.db 0x55,0x55,0x55,0x55,0x55,0x55,0x55,0x55,0x55,0x55	;21.-140. Pixel, cyan
.db 0x66,0x66,0x66,0x66,0x66,0x66,0x66,0x66,0x66,0x66	;21.-160. Pixel, rotblau
.db 0x70,0x70,0x70,0x70,0x70,0x70,0x70,0x70,0x70,0x70	;21.-180. Pixel, grau
.db  0x00,0x00,0x00,0x00,0x00,0x0F			;181.-192. Pixel, schwarz
;; 50. Zeile:					
.db 0xF0,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00	;erste 10 Bytes = 20 Pixel, schwarz
.db 0x11,0x11,0x11,0x11,0x11,0x11,0x11,0x11,0x11,0x11	;21.-40. Pixel, rot
.db 0x22,0x22,0x22,0x22,0x22,0x22,0x22,0x22,0x22,0x22	;41.-60. Pixel, gruen
.db 0x44,0x44,0x44,0x44,0x44,0x44,0x44,0x44,0x44,0x44	;61.-80. Pixel, blau
.db 0x77,0x77,0x77,0x77,0x77,0x77,0x77,0x77,0x77,0x77	;21.-100. Pixel, weiss
.db 0x33,0x33,0x33,0x33,0x33,0x33,0x33,0x33,0x33,0x33	;21.-120. Pixel, gelb
.db 0x55,0x55,0x55,0x55,0x55,0x55,0x55,0x55,0x55,0x55	;21.-140. Pixel, cyan
.db 0x66,0x66,0x66,0x66,0x66,0x66,0x66,0x66,0x66,0x66	;21.-160. Pixel, rotblau
.db 0x70,0x70,0x70,0x70,0x70,0x70,0x70,0x70,0x70,0x70	;21.-180. Pixel, grau
.db  0x00,0x00,0x00,0x00,0x00,0x0F			;181.-192. Pixel, schwarz
;; 11.-20. Zeile:
.db 0xF0,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00	;erste 10 Bytes = 20 Pixel, schwarz
.db 0x11,0x11,0x11,0x11,0x11,0x11,0x11,0x11,0x11,0x11	;21.-40. Pixel, rot
.db 0x22,0x22,0x22,0x22,0x22,0x22,0x22,0x22,0x22,0x22	;41.-60. Pixel, gruen
.db 0x44,0x44,0x44,0x44,0x44,0x44,0x44,0x44,0x44,0x44	;61.-80. Pixel, blau
.db 0x77,0x77,0x77,0x77,0x77,0x77,0x77,0x77,0x77,0x77	;21.-100. Pixel, weiss
.db 0x33,0x33,0x33,0x33,0x33,0x33,0x33,0x33,0x33,0x33	;21.-120. Pixel, gelb
.db 0x55,0x55,0x55,0x55,0x55,0x55,0x55,0x55,0x55,0x55	;21.-140. Pixel, cyan
.db 0x66,0x66,0x66,0x66,0x66,0x66,0x66,0x66,0x66,0x66	;21.-160. Pixel, rotblau
.db 0x70,0x70,0x70,0x70,0x70,0x70,0x70,0x70,0x70,0x70	;21.-180. Pixel, grau
.db  0x00,0x00,0x00,0x00,0x00,0x0F			;181.-192. Pixel, schwarz
;; 12. Zeile:					
.db 0xF0,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00	;erste 10 Bytes = 20 Pixel, schwarz
.db 0x11,0x11,0x11,0x11,0x11,0x11,0x11,0x11,0x11,0x11	;21.-40. Pixel, rot
.db 0x22,0x22,0x22,0x22,0x22,0x22,0x22,0x22,0x22,0x22	;41.-60. Pixel, gruen
.db 0x44,0x44,0x44,0x44,0x44,0x44,0x44,0x44,0x44,0x44	;61.-80. Pixel, blau
.db 0x77,0x77,0x77,0x77,0x77,0x77,0x77,0x77,0x77,0x77	;21.-100. Pixel, weiss
.db 0x33,0x33,0x33,0x33,0x33,0x33,0x33,0x33,0x33,0x33	;21.-120. Pixel, gelb
.db 0x55,0x55,0x55,0x55,0x55,0x55,0x55,0x55,0x55,0x55	;21.-140. Pixel, cyan
.db 0x66,0x66,0x66,0x66,0x66,0x66,0x66,0x66,0x66,0x66	;21.-160. Pixel, rotblau
.db 0x70,0x70,0x70,0x70,0x70,0x70,0x70,0x70,0x70,0x70	;21.-180. Pixel, grau
.db  0x00,0x00,0x00,0x00,0x00,0x0F			;181.-192. Pixel, schwarz
;; 13. Zeile:					
.db 0xF0,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00	;erste 10 Bytes = 20 Pixel, schwarz
.db 0x11,0x11,0x11,0x11,0x11,0x11,0x11,0x11,0x11,0x11	;21.-40. Pixel, rot
.db 0x22,0x22,0x22,0x22,0x22,0x22,0x22,0x22,0x22,0x22	;41.-60. Pixel, gruen
.db 0x44,0x44,0x44,0x44,0x44,0x44,0x44,0x44,0x44,0x44	;61.-80. Pixel, blau
.db 0x77,0x77,0x77,0x77,0x77,0x77,0x77,0x77,0x77,0x77	;21.-100. Pixel, weiss
.db 0x33,0x33,0x33,0x33,0x33,0x33,0x33,0x33,0x33,0x33	;21.-120. Pixel, gelb
.db 0x55,0x55,0x55,0x55,0x55,0x55,0x55,0x55,0x55,0x55	;21.-140. Pixel, cyan
.db 0x66,0x66,0x66,0x66,0x66,0x66,0x66,0x66,0x66,0x66	;21.-160. Pixel, rotblau
.db 0x70,0x70,0x70,0x70,0x70,0x70,0x70,0x70,0x70,0x70	;21.-180. Pixel, grau
.db  0x00,0x00,0x00,0x00,0x00,0x0F			;181.-192. Pixel, schwarz
;; 14. Zeile:					
.db 0xF0,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00	;erste 10 Bytes = 20 Pixel, schwarz
.db 0x11,0x11,0x11,0x11,0x11,0x11,0x11,0x11,0x11,0x11	;21.-40. Pixel, rot
.db 0x22,0x22,0x22,0x22,0x22,0x22,0x22,0x22,0x22,0x22	;41.-60. Pixel, gruen
.db 0x44,0x44,0x44,0x44,0x44,0x44,0x44,0x44,0x44,0x44	;61.-80. Pixel, blau
.db 0x77,0x77,0x77,0x77,0x77,0x77,0x77,0x77,0x77,0x77	;21.-100. Pixel, weiss
.db 0x33,0x33,0x33,0x33,0x33,0x33,0x33,0x33,0x33,0x33	;21.-120. Pixel, gelb
.db 0x55,0x55,0x55,0x55,0x55,0x55,0x55,0x55,0x55,0x55	;21.-140. Pixel, cyan
.db 0x66,0x66,0x66,0x66,0x66,0x66,0x66,0x66,0x66,0x66	;21.-160. Pixel, rotblau
.db 0x70,0x70,0x70,0x70,0x70,0x70,0x70,0x70,0x70,0x70	;21.-180. Pixel, grau
.db  0x00,0x00,0x00,0x00,0x00,0x0F			;181.-192. Pixel, schwarz
;; 15. Zeile:					
.db 0xF0,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00	;erste 10 Bytes = 20 Pixel, schwarz
.db 0x11,0x11,0x11,0x11,0x11,0x11,0x11,0x11,0x11,0x11	;21.-40. Pixel, rot
.db 0x22,0x22,0x22,0x22,0x22,0x22,0x22,0x22,0x22,0x22	;41.-60. Pixel, gruen
.db 0x44,0x44,0x44,0x44,0x44,0x44,0x44,0x44,0x44,0x44	;61.-80. Pixel, blau
.db 0x77,0x77,0x77,0x77,0x77,0x77,0x77,0x77,0x77,0x77	;21.-100. Pixel, weiss
.db 0x33,0x33,0x33,0x33,0x33,0x33,0x33,0x33,0x33,0x33	;21.-120. Pixel, gelb
.db 0x55,0x55,0x55,0x55,0x55,0x55,0x55,0x55,0x55,0x55	;21.-140. Pixel, cyan
.db 0x66,0x66,0x66,0x66,0x66,0x66,0x66,0x66,0x66,0x66	;21.-160. Pixel, rotblau
.db 0x70,0x70,0x70,0x70,0x70,0x70,0x70,0x70,0x70,0x70	;21.-180. Pixel, grau
.db  0x00,0x00,0x00,0x00,0x00,0x0F			;181.-192. Pixel, schwarz
;; 16. Zeile:					
.db 0xF0,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00	;erste 10 Bytes = 20 Pixel, schwarz
.db 0x11,0x11,0x11,0x11,0x11,0x11,0x11,0x11,0x11,0x11	;21.-40. Pixel, rot
.db 0x22,0x22,0x22,0x22,0x22,0x22,0x22,0x22,0x22,0x22	;41.-60. Pixel, gruen
.db 0x44,0x44,0x44,0x44,0x44,0x44,0x44,0x44,0x44,0x44	;61.-80. Pixel, blau
.db 0x77,0x77,0x77,0x77,0x77,0x77,0x77,0x77,0x77,0x77	;21.-100. Pixel, weiss
.db 0x33,0x33,0x33,0x33,0x33,0x33,0x33,0x33,0x33,0x33	;21.-120. Pixel, gelb
.db 0x55,0x55,0x55,0x55,0x55,0x55,0x55,0x55,0x55,0x55	;21.-140. Pixel, cyan
.db 0x66,0x66,0x66,0x66,0x66,0x66,0x66,0x66,0x66,0x66	;21.-160. Pixel, rotblau
.db 0x70,0x70,0x70,0x70,0x70,0x70,0x70,0x70,0x70,0x70	;21.-180. Pixel, grau
.db  0x00,0x00,0x00,0x00,0x00,0x0F			;181.-192. Pixel, schwarz
;; 17. Zeile:					
.db 0xF0,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00	;erste 10 Bytes = 20 Pixel, schwarz
.db 0x11,0x11,0x11,0x11,0x11,0x11,0x11,0x11,0x11,0x11	;21.-40. Pixel, rot
.db 0x22,0x22,0x22,0x22,0x22,0x22,0x22,0x22,0x22,0x22	;41.-60. Pixel, gruen
.db 0x44,0x44,0x44,0x44,0x44,0x44,0x44,0x44,0x44,0x44	;61.-80. Pixel, blau
.db 0x77,0x77,0x77,0x77,0x77,0x77,0x77,0x77,0x77,0x77	;21.-100. Pixel, weiss
.db 0x33,0x33,0x33,0x33,0x33,0x33,0x33,0x33,0x33,0x33	;21.-120. Pixel, gelb
.db 0x55,0x55,0x55,0x55,0x55,0x55,0x55,0x55,0x55,0x55	;21.-140. Pixel, cyan
.db 0x66,0x66,0x66,0x66,0x66,0x66,0x66,0x66,0x66,0x66	;21.-160. Pixel, rotblau
.db 0x70,0x70,0x70,0x70,0x70,0x70,0x70,0x70,0x70,0x70	;21.-180. Pixel, grau
.db  0x00,0x00,0x00,0x00,0x00,0x0F			;181.-192. Pixel, schwarz
;; 18. Zeile:					
.db 0xF0,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00	;erste 10 Bytes = 20 Pixel, schwarz
.db 0x11,0x11,0x11,0x11,0x11,0x11,0x11,0x11,0x11,0x11	;21.-40. Pixel, rot
.db 0x22,0x22,0x22,0x22,0x22,0x22,0x22,0x22,0x22,0x22	;41.-60. Pixel, gruen
.db 0x44,0x44,0x44,0x44,0x44,0x44,0x44,0x44,0x44,0x44	;61.-80. Pixel, blau
.db 0x77,0x77,0x77,0x77,0x77,0x77,0x77,0x77,0x77,0x77	;21.-100. Pixel, weiss
.db 0x33,0x33,0x33,0x33,0x33,0x33,0x33,0x33,0x33,0x33	;21.-120. Pixel, gelb
.db 0x55,0x55,0x55,0x55,0x55,0x55,0x55,0x55,0x55,0x55	;21.-140. Pixel, cyan
.db 0x66,0x66,0x66,0x66,0x66,0x66,0x66,0x66,0x66,0x66	;21.-160. Pixel, rotblau
.db 0x70,0x70,0x70,0x70,0x70,0x70,0x70,0x70,0x70,0x70	;21.-180. Pixel, grau
.db  0x00,0x00,0x00,0x00,0x00,0x0F			;181.-192. Pixel, schwarz
;; 19. Zeile:					
.db 0xF0,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00	;erste 10 Bytes = 20 Pixel, schwarz
.db 0x11,0x11,0x11,0x11,0x11,0x11,0x11,0x11,0x11,0x11	;21.-40. Pixel, rot
.db 0x22,0x22,0x22,0x22,0x22,0x22,0x22,0x22,0x22,0x22	;41.-60. Pixel, gruen
.db 0x44,0x44,0x44,0x44,0x44,0x44,0x44,0x44,0x44,0x44	;61.-80. Pixel, blau
.db 0x77,0x77,0x77,0x77,0x77,0x77,0x77,0x77,0x77,0x77	;21.-100. Pixel, weiss
.db 0x33,0x33,0x33,0x33,0x33,0x33,0x33,0x33,0x33,0x33	;21.-120. Pixel, gelb
.db 0x55,0x55,0x55,0x55,0x55,0x55,0x55,0x55,0x55,0x55	;21.-140. Pixel, cyan
.db 0x66,0x66,0x66,0x66,0x66,0x66,0x66,0x66,0x66,0x66	;21.-160. Pixel, rotblau
.db 0x70,0x70,0x70,0x70,0x70,0x70,0x70,0x70,0x70,0x70	;21.-180. Pixel, grau
.db  0x00,0x00,0x00,0x00,0x00,0x0F			;181.-192. Pixel, schwarz
;; 60. Zeile:
.db 0xF0,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00	;erste 10 Bytes = 20 Pixel, schwarz
.db 0x11,0x11,0x11,0x11,0x11,0x11,0x11,0x11,0x11,0x11	;21.-40. Pixel, rot
.db 0x22,0x22,0x22,0x22,0x22,0x22,0x22,0x22,0x22,0x22	;41.-60. Pixel, gruen
.db 0x44,0x44,0x44,0x44,0x44,0x44,0x44,0x44,0x44,0x44	;61.-80. Pixel, blau
.db 0x77,0x77,0x77,0x77,0x77,0x77,0x77,0x77,0x77,0x77	;21.-100. Pixel, weiss
.db 0x33,0x33,0x33,0x33,0x33,0x33,0x33,0x33,0x33,0x33	;21.-120. Pixel, gelb
.db 0x55,0x55,0x55,0x55,0x55,0x55,0x55,0x55,0x55,0x55	;21.-140. Pixel, cyan
.db 0x66,0x66,0x66,0x66,0x66,0x66,0x66,0x66,0x66,0x66	;21.-160. Pixel, rotblau
.db 0x70,0x70,0x70,0x70,0x70,0x70,0x70,0x70,0x70,0x70	;21.-180. Pixel, grau
.db  0x00,0x00,0x00,0x00,0x00,0x0F			;181.-192. Pixel, schwarz
;; 21.-40. Zeile:
.db 0xF0,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00	;erste 10 Bytes = 20 Pixel, schwarz
.db 0x11,0x11,0x11,0x11,0x11,0x11,0x11,0x11,0x11,0x11	;21.-40. Pixel, rot
.db 0x22,0x22,0x22,0x22,0x22,0x22,0x22,0x22,0x22,0x22	;41.-60. Pixel, gruen
.db 0x44,0x44,0x44,0x44,0x44,0x44,0x44,0x44,0x44,0x44	;61.-80. Pixel, blau
.db 0x77,0x77,0x77,0x77,0x77,0x77,0x77,0x77,0x77,0x77	;21.-100. Pixel, weiss
.db 0x33,0x33,0x33,0x33,0x33,0x33,0x33,0x33,0x33,0x33	;21.-120. Pixel, gelb
.db 0x55,0x55,0x55,0x55,0x55,0x55,0x55,0x55,0x55,0x55	;21.-140. Pixel, cyan
.db 0x66,0x66,0x66,0x66,0x66,0x66,0x66,0x66,0x66,0x66	;21.-160. Pixel, rotblau
.db 0x70,0x70,0x70,0x70,0x70,0x70,0x70,0x70,0x70,0x70	;21.-180. Pixel, grau
.db  0x00,0x00,0x00,0x00,0x00,0x0F			;181.-192. Pixel, schwarz
;; 2. Zeile:					
.db 0xF0,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00	;erste 10 Bytes = 20 Pixel, schwarz
.db 0x11,0x11,0x11,0x11,0x11,0x11,0x11,0x11,0x11,0x11	;21.-40. Pixel, rot
.db 0x22,0x22,0x22,0x22,0x22,0x22,0x22,0x22,0x22,0x22	;41.-60. Pixel, gruen
.db 0x44,0x44,0x44,0x44,0x44,0x44,0x44,0x44,0x44,0x44	;61.-80. Pixel, blau
.db 0x77,0x77,0x77,0x77,0x77,0x77,0x77,0x77,0x77,0x77	;21.-100. Pixel, weiss
.db 0x33,0x33,0x33,0x33,0x33,0x33,0x33,0x33,0x33,0x33	;21.-120. Pixel, gelb
.db 0x55,0x55,0x55,0x55,0x55,0x55,0x55,0x55,0x55,0x55	;21.-140. Pixel, cyan
.db 0x66,0x66,0x66,0x66,0x66,0x66,0x66,0x66,0x66,0x66	;21.-160. Pixel, rotblau
.db 0x70,0x70,0x70,0x70,0x70,0x70,0x70,0x70,0x70,0x70	;21.-180. Pixel, grau
.db  0x00,0x00,0x00,0x00,0x00,0x0F			;181.-192. Pixel, schwarz
;; 3. Zeile:					
.db 0xF0,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00	;erste 10 Bytes = 20 Pixel, schwarz
.db 0x11,0x11,0x11,0x11,0x11,0x11,0x11,0x11,0x11,0x11	;21.-40. Pixel, rot
.db 0x22,0x22,0x22,0x22,0x22,0x22,0x22,0x22,0x22,0x22	;41.-60. Pixel, gruen
.db 0x44,0x44,0x44,0x44,0x44,0x44,0x44,0x44,0x44,0x44	;61.-80. Pixel, blau
.db 0x77,0x77,0x77,0x77,0x77,0x77,0x77,0x77,0x77,0x77	;21.-100. Pixel, weiss
.db 0x33,0x33,0x33,0x33,0x33,0x33,0x33,0x33,0x33,0x33	;21.-120. Pixel, gelb
.db 0x55,0x55,0x55,0x55,0x55,0x55,0x55,0x55,0x55,0x55	;21.-140. Pixel, cyan
.db 0x66,0x66,0x66,0x66,0x66,0x66,0x66,0x66,0x66,0x66	;21.-160. Pixel, rotblau
.db 0x70,0x70,0x70,0x70,0x70,0x70,0x70,0x70,0x70,0x70	;21.-180. Pixel, grau
.db  0x00,0x00,0x00,0x00,0x00,0x0F			;181.-192. Pixel, schwarz
;; 4. Zeile:					
.db 0xF0,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00	;erste 10 Bytes = 20 Pixel, schwarz
.db 0x11,0x11,0x11,0x11,0x11,0x11,0x11,0x11,0x11,0x11	;21.-40. Pixel, rot
.db 0x22,0x22,0x22,0x22,0x22,0x22,0x22,0x22,0x22,0x22	;41.-60. Pixel, gruen
.db 0x44,0x44,0x44,0x44,0x44,0x44,0x44,0x44,0x44,0x44	;61.-80. Pixel, blau
.db 0x77,0x77,0x77,0x77,0x77,0x77,0x77,0x77,0x77,0x77	;21.-100. Pixel, weiss
.db 0x33,0x33,0x33,0x33,0x33,0x33,0x33,0x33,0x33,0x33	;21.-120. Pixel, gelb
.db 0x55,0x55,0x55,0x55,0x55,0x55,0x55,0x55,0x55,0x55	;21.-140. Pixel, cyan
.db 0x66,0x66,0x66,0x66,0x66,0x66,0x66,0x66,0x66,0x66	;21.-160. Pixel, rotblau
.db 0x70,0x70,0x70,0x70,0x70,0x70,0x70,0x70,0x70,0x70	;21.-180. Pixel, grau
.db  0x00,0x00,0x00,0x00,0x00,0x0F			;181.-192. Pixel, schwarz
;; 5. Zeile:					
.db 0xF0,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00	;erste 10 Bytes = 20 Pixel, schwarz
.db 0x11,0x11,0x11,0x11,0x11,0x11,0x11,0x11,0x11,0x11	;21.-40. Pixel, rot
.db 0x22,0x22,0x22,0x22,0x22,0x22,0x22,0x22,0x22,0x22	;41.-60. Pixel, gruen
.db 0x44,0x44,0x44,0x44,0x44,0x44,0x44,0x44,0x44,0x44	;61.-80. Pixel, blau
.db 0x77,0x77,0x77,0x77,0x77,0x77,0x77,0x77,0x77,0x77	;21.-100. Pixel, weiss
.db 0x33,0x33,0x33,0x33,0x33,0x33,0x33,0x33,0x33,0x33	;21.-120. Pixel, gelb
.db 0x55,0x55,0x55,0x55,0x55,0x55,0x55,0x55,0x55,0x55	;21.-140. Pixel, cyan
.db 0x66,0x66,0x66,0x66,0x66,0x66,0x66,0x66,0x66,0x66	;21.-160. Pixel, rotblau
.db 0x70,0x70,0x70,0x70,0x70,0x70,0x70,0x70,0x70,0x70	;21.-180. Pixel, grau
.db  0x00,0x00,0x00,0x00,0x00,0x0F			;181.-192. Pixel, schwarz
;; 6. Zeile:					
.db 0xF0,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00	;erste 10 Bytes = 20 Pixel, schwarz
.db 0x11,0x11,0x11,0x11,0x11,0x11,0x11,0x11,0x11,0x11	;21.-40. Pixel, rot
.db 0x22,0x22,0x22,0x22,0x22,0x22,0x22,0x22,0x22,0x22	;41.-60. Pixel, gruen
.db 0x44,0x44,0x44,0x44,0x44,0x44,0x44,0x44,0x44,0x44	;61.-80. Pixel, blau
.db 0x77,0x77,0x77,0x77,0x77,0x77,0x77,0x77,0x77,0x77	;21.-100. Pixel, weiss
.db 0x33,0x33,0x33,0x33,0x33,0x33,0x33,0x33,0x33,0x33	;21.-120. Pixel, gelb
.db 0x55,0x55,0x55,0x55,0x55,0x55,0x55,0x55,0x55,0x55	;21.-140. Pixel, cyan
.db 0x66,0x66,0x66,0x66,0x66,0x66,0x66,0x66,0x66,0x66	;21.-160. Pixel, rotblau
.db 0x70,0x70,0x70,0x70,0x70,0x70,0x70,0x70,0x70,0x70	;21.-180. Pixel, grau
.db  0x00,0x00,0x00,0x00,0x00,0x0F			;181.-192. Pixel, schwarz
;; 7. Zeile:					
.db 0xF0,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00	;erste 10 Bytes = 20 Pixel, schwarz
.db 0x11,0x11,0x11,0x11,0x11,0x11,0x11,0x11,0x11,0x11	;21.-40. Pixel, rot
.db 0x22,0x22,0x22,0x22,0x22,0x22,0x22,0x22,0x22,0x22	;41.-60. Pixel, gruen
.db 0x44,0x44,0x44,0x44,0x44,0x44,0x44,0x44,0x44,0x44	;61.-80. Pixel, blau
.db 0x77,0x77,0x77,0x77,0x77,0x77,0x77,0x77,0x77,0x77	;21.-100. Pixel, weiss
.db 0x33,0x33,0x33,0x33,0x33,0x33,0x33,0x33,0x33,0x33	;21.-120. Pixel, gelb
.db 0x55,0x55,0x55,0x55,0x55,0x55,0x55,0x55,0x55,0x55	;21.-140. Pixel, cyan
.db 0x66,0x66,0x66,0x66,0x66,0x66,0x66,0x66,0x66,0x66	;21.-160. Pixel, rotblau
.db 0x70,0x70,0x70,0x70,0x70,0x70,0x70,0x70,0x70,0x70	;21.-180. Pixel, grau
.db  0x00,0x00,0x00,0x00,0x00,0x0F			;181.-192. Pixel, schwarz
;; 8. Zeile:					
.db 0xF0,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00	;erste 10 Bytes = 20 Pixel, schwarz
.db 0x11,0x11,0x11,0x11,0x11,0x11,0x11,0x11,0x11,0x11	;21.-40. Pixel, rot
.db 0x22,0x22,0x22,0x22,0x22,0x22,0x22,0x22,0x22,0x22	;41.-60. Pixel, gruen
.db 0x44,0x44,0x44,0x44,0x44,0x44,0x44,0x44,0x44,0x44	;61.-80. Pixel, blau
.db 0x77,0x77,0x77,0x77,0x77,0x77,0x77,0x77,0x77,0x77	;21.-100. Pixel, weiss
.db 0x33,0x33,0x33,0x33,0x33,0x33,0x33,0x33,0x33,0x33	;21.-120. Pixel, gelb
.db 0x55,0x55,0x55,0x55,0x55,0x55,0x55,0x55,0x55,0x55	;21.-140. Pixel, cyan
.db 0x66,0x66,0x66,0x66,0x66,0x66,0x66,0x66,0x66,0x66	;21.-160. Pixel, rotblau
.db 0x70,0x70,0x70,0x70,0x70,0x70,0x70,0x70,0x70,0x70	;21.-180. Pixel, grau
.db  0x00,0x00,0x00,0x00,0x00,0x0F			;181.-192. Pixel, schwarz
;; 9. Zeile:					
.db 0xF0,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00	;erste 10 Bytes = 20 Pixel, schwarz
.db 0x11,0x11,0x11,0x11,0x11,0x11,0x11,0x11,0x11,0x11	;21.-40. Pixel, rot
.db 0x22,0x22,0x22,0x22,0x22,0x22,0x22,0x22,0x22,0x22	;41.-60. Pixel, gruen
.db 0x44,0x44,0x44,0x44,0x44,0x44,0x44,0x44,0x44,0x44	;61.-80. Pixel, blau
.db 0x77,0x77,0x77,0x77,0x77,0x77,0x77,0x77,0x77,0x77	;21.-100. Pixel, weiss
.db 0x33,0x33,0x33,0x33,0x33,0x33,0x33,0x33,0x33,0x33	;21.-120. Pixel, gelb
.db 0x55,0x55,0x55,0x55,0x55,0x55,0x55,0x55,0x55,0x55	;21.-140. Pixel, cyan
.db 0x66,0x66,0x66,0x66,0x66,0x66,0x66,0x66,0x66,0x66	;21.-160. Pixel, rotblau
.db 0x70,0x70,0x70,0x70,0x70,0x70,0x70,0x70,0x70,0x70	;21.-180. Pixel, grau
.db  0x00,0x00,0x00,0x00,0x00,0x0F			;181.-192. Pixel, schwarz
;; 50. Zeile:					
.db 0xF0,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00	;erste 10 Bytes = 20 Pixel, schwarz
.db 0x11,0x11,0x11,0x11,0x11,0x11,0x11,0x11,0x11,0x11	;21.-40. Pixel, rot
.db 0x22,0x22,0x22,0x22,0x22,0x22,0x22,0x22,0x22,0x22	;41.-60. Pixel, gruen
.db 0x44,0x44,0x44,0x44,0x44,0x44,0x44,0x44,0x44,0x44	;61.-80. Pixel, blau
.db 0x77,0x77,0x77,0x77,0x77,0x77,0x77,0x77,0x77,0x77	;21.-100. Pixel, weiss
.db 0x33,0x33,0x33,0x33,0x33,0x33,0x33,0x33,0x33,0x33	;21.-120. Pixel, gelb
.db 0x55,0x55,0x55,0x55,0x55,0x55,0x55,0x55,0x55,0x55	;21.-140. Pixel, cyan
.db 0x66,0x66,0x66,0x66,0x66,0x66,0x66,0x66,0x66,0x66	;21.-160. Pixel, rotblau
.db 0x70,0x70,0x70,0x70,0x70,0x70,0x70,0x70,0x70,0x70	;21.-180. Pixel, grau
.db  0x00,0x00,0x00,0x00,0x00,0x0F			;181.-192. Pixel, schwarz
;; 11.-20. Zeile:
.db 0xF0,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00	;erste 10 Bytes = 20 Pixel, schwarz
.db 0x11,0x11,0x11,0x11,0x11,0x11,0x11,0x11,0x11,0x11	;21.-40. Pixel, rot
.db 0x22,0x22,0x22,0x22,0x22,0x22,0x22,0x22,0x22,0x22	;41.-60. Pixel, gruen
.db 0x44,0x44,0x44,0x44,0x44,0x44,0x44,0x44,0x44,0x44	;61.-80. Pixel, blau
.db 0x77,0x77,0x77,0x77,0x77,0x77,0x77,0x77,0x77,0x77	;21.-100. Pixel, weiss
.db 0x33,0x33,0x33,0x33,0x33,0x33,0x33,0x33,0x33,0x33	;21.-120. Pixel, gelb
.db 0x55,0x55,0x55,0x55,0x55,0x55,0x55,0x55,0x55,0x55	;21.-140. Pixel, cyan
.db 0x66,0x66,0x66,0x66,0x66,0x66,0x66,0x66,0x66,0x66	;21.-160. Pixel, rotblau
.db 0x70,0x70,0x70,0x70,0x70,0x70,0x70,0x70,0x70,0x70	;21.-180. Pixel, grau
.db  0x00,0x00,0x00,0x00,0x00,0x0F			;181.-192. Pixel, schwarz
;; 12. Zeile:					
.db 0xF0,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00	;erste 10 Bytes = 20 Pixel, schwarz
.db 0x11,0x11,0x11,0x11,0x11,0x11,0x11,0x11,0x11,0x11	;21.-40. Pixel, rot
.db 0x22,0x22,0x22,0x22,0x22,0x22,0x22,0x22,0x22,0x22	;41.-60. Pixel, gruen
.db 0x44,0x44,0x44,0x44,0x44,0x44,0x44,0x44,0x44,0x44	;61.-80. Pixel, blau
.db 0x77,0x77,0x77,0x77,0x77,0x77,0x77,0x77,0x77,0x77	;21.-100. Pixel, weiss
.db 0x33,0x33,0x33,0x33,0x33,0x33,0x33,0x33,0x33,0x33	;21.-120. Pixel, gelb
.db 0x55,0x55,0x55,0x55,0x55,0x55,0x55,0x55,0x55,0x55	;21.-140. Pixel, cyan
.db 0x66,0x66,0x66,0x66,0x66,0x66,0x66,0x66,0x66,0x66	;21.-160. Pixel, rotblau
.db 0x70,0x70,0x70,0x70,0x70,0x70,0x70,0x70,0x70,0x70	;21.-180. Pixel, grau
.db  0x00,0x00,0x00,0x00,0x00,0x0F			;181.-192. Pixel, schwarz
;; 13. Zeile:					
.db 0xF0,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00	;erste 10 Bytes = 20 Pixel, schwarz
.db 0x11,0x11,0x11,0x11,0x11,0x11,0x11,0x11,0x11,0x11	;21.-40. Pixel, rot
.db 0x22,0x22,0x22,0x22,0x22,0x22,0x22,0x22,0x22,0x22	;41.-60. Pixel, gruen
.db 0x44,0x44,0x44,0x44,0x44,0x44,0x44,0x44,0x44,0x44	;61.-80. Pixel, blau
.db 0x77,0x77,0x77,0x77,0x77,0x77,0x77,0x77,0x77,0x77	;21.-100. Pixel, weiss
.db 0x33,0x33,0x33,0x33,0x33,0x33,0x33,0x33,0x33,0x33	;21.-120. Pixel, gelb
.db 0x55,0x55,0x55,0x55,0x55,0x55,0x55,0x55,0x55,0x55	;21.-140. Pixel, cyan
.db 0x66,0x66,0x66,0x66,0x66,0x66,0x66,0x66,0x66,0x66	;21.-160. Pixel, rotblau
.db 0x70,0x70,0x70,0x70,0x70,0x70,0x70,0x70,0x70,0x70	;21.-180. Pixel, grau
.db  0x00,0x00,0x00,0x00,0x00,0x0F			;181.-192. Pixel, schwarz
;; 14. Zeile:					
.db 0xF0,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00	;erste 10 Bytes = 20 Pixel, schwarz
.db 0x11,0x11,0x11,0x11,0x11,0x11,0x11,0x11,0x11,0x11	;21.-40. Pixel, rot
.db 0x22,0x22,0x22,0x22,0x22,0x22,0x22,0x22,0x22,0x22	;41.-60. Pixel, gruen
.db 0x44,0x44,0x44,0x44,0x44,0x44,0x44,0x44,0x44,0x44	;61.-80. Pixel, blau
.db 0x77,0x77,0x77,0x77,0x77,0x77,0x77,0x77,0x77,0x77	;21.-100. Pixel, weiss
.db 0x33,0x33,0x33,0x33,0x33,0x33,0x33,0x33,0x33,0x33	;21.-120. Pixel, gelb
.db 0x55,0x55,0x55,0x55,0x55,0x55,0x55,0x55,0x55,0x55	;21.-140. Pixel, cyan
.db 0x66,0x66,0x66,0x66,0x66,0x66,0x66,0x66,0x66,0x66	;21.-160. Pixel, rotblau
.db 0x70,0x70,0x70,0x70,0x70,0x70,0x70,0x70,0x70,0x70	;21.-180. Pixel, grau
.db  0x00,0x00,0x00,0x00,0x00,0x0F			;181.-192. Pixel, schwarz
;; 15. Zeile:					
.db 0xF0,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00	;erste 10 Bytes = 20 Pixel, schwarz
.db 0x11,0x11,0x11,0x11,0x11,0x11,0x11,0x11,0x11,0x11	;21.-40. Pixel, rot
.db 0x22,0x22,0x22,0x22,0x22,0x22,0x22,0x22,0x22,0x22	;41.-60. Pixel, gruen
.db 0x44,0x44,0x44,0x44,0x44,0x44,0x44,0x44,0x44,0x44	;61.-80. Pixel, blau
.db 0x77,0x77,0x77,0x77,0x77,0x77,0x77,0x77,0x77,0x77	;21.-100. Pixel, weiss
.db 0x33,0x33,0x33,0x33,0x33,0x33,0x33,0x33,0x33,0x33	;21.-120. Pixel, gelb
.db 0x55,0x55,0x55,0x55,0x55,0x55,0x55,0x55,0x55,0x55	;21.-140. Pixel, cyan
.db 0x66,0x66,0x66,0x66,0x66,0x66,0x66,0x66,0x66,0x66	;21.-160. Pixel, rotblau
.db 0x70,0x70,0x70,0x70,0x70,0x70,0x70,0x70,0x70,0x70	;21.-180. Pixel, grau
.db  0x00,0x00,0x00,0x00,0x00,0x0F			;181.-192. Pixel, schwarz
;; 16. Zeile:					
.db 0xF0,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00	;erste 10 Bytes = 20 Pixel, schwarz
.db 0x11,0x11,0x11,0x11,0x11,0x11,0x11,0x11,0x11,0x11	;21.-40. Pixel, rot
.db 0x22,0x22,0x22,0x22,0x22,0x22,0x22,0x22,0x22,0x22	;41.-60. Pixel, gruen
.db 0x44,0x44,0x44,0x44,0x44,0x44,0x44,0x44,0x44,0x44	;61.-80. Pixel, blau
.db 0x77,0x77,0x77,0x77,0x77,0x77,0x77,0x77,0x77,0x77	;21.-100. Pixel, weiss
.db 0x33,0x33,0x33,0x33,0x33,0x33,0x33,0x33,0x33,0x33	;21.-120. Pixel, gelb
.db 0x55,0x55,0x55,0x55,0x55,0x55,0x55,0x55,0x55,0x55	;21.-140. Pixel, cyan
.db 0x66,0x66,0x66,0x66,0x66,0x66,0x66,0x66,0x66,0x66	;21.-160. Pixel, rotblau
.db 0x70,0x70,0x70,0x70,0x70,0x70,0x70,0x70,0x70,0x70	;21.-180. Pixel, grau
.db  0x00,0x00,0x00,0x00,0x00,0x0F			;181.-192. Pixel, schwarz
;; 17. Zeile:					
.db 0xF0,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00	;erste 10 Bytes = 20 Pixel, schwarz
.db 0x11,0x11,0x11,0x11,0x11,0x11,0x11,0x11,0x11,0x11	;21.-40. Pixel, rot
.db 0x22,0x22,0x22,0x22,0x22,0x22,0x22,0x22,0x22,0x22	;41.-60. Pixel, gruen
.db 0x44,0x44,0x44,0x44,0x44,0x44,0x44,0x44,0x44,0x44	;61.-80. Pixel, blau
.db 0x77,0x77,0x77,0x77,0x77,0x77,0x77,0x77,0x77,0x77	;21.-100. Pixel, weiss
.db 0x33,0x33,0x33,0x33,0x33,0x33,0x33,0x33,0x33,0x33	;21.-120. Pixel, gelb
.db 0x55,0x55,0x55,0x55,0x55,0x55,0x55,0x55,0x55,0x55	;21.-140. Pixel, cyan
.db 0x66,0x66,0x66,0x66,0x66,0x66,0x66,0x66,0x66,0x66	;21.-160. Pixel, rotblau
.db 0x70,0x70,0x70,0x70,0x70,0x70,0x70,0x70,0x70,0x70	;21.-180. Pixel, grau
.db  0x00,0x00,0x00,0x00,0x00,0x0F			;181.-192. Pixel, schwarz
;; 18. Zeile:					
.db 0xF0,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00	;erste 10 Bytes = 20 Pixel, schwarz
.db 0x11,0x11,0x11,0x11,0x11,0x11,0x11,0x11,0x11,0x11	;21.-40. Pixel, rot
.db 0x22,0x22,0x22,0x22,0x22,0x22,0x22,0x22,0x22,0x22	;41.-60. Pixel, gruen
.db 0x44,0x44,0x44,0x44,0x44,0x44,0x44,0x44,0x44,0x44	;61.-80. Pixel, blau
.db 0x77,0x77,0x77,0x77,0x77,0x77,0x77,0x77,0x77,0x77	;21.-100. Pixel, weiss
.db 0x33,0x33,0x33,0x33,0x33,0x33,0x33,0x33,0x33,0x33	;21.-120. Pixel, gelb
.db 0x55,0x55,0x55,0x55,0x55,0x55,0x55,0x55,0x55,0x55	;21.-140. Pixel, cyan
.db 0x66,0x66,0x66,0x66,0x66,0x66,0x66,0x66,0x66,0x66	;21.-160. Pixel, rotblau
.db 0x70,0x70,0x70,0x70,0x70,0x70,0x70,0x70,0x70,0x70	;21.-180. Pixel, grau
.db  0x00,0x00,0x00,0x00,0x00,0x0F			;181.-192. Pixel, schwarz
;; 19. Zeile:					
.db 0xF0,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00	;erste 10 Bytes = 20 Pixel, schwarz
.db 0x11,0x11,0x11,0x11,0x11,0x11,0x11,0x11,0x11,0x11	;21.-40. Pixel, rot
.db 0x22,0x22,0x22,0x22,0x22,0x22,0x22,0x22,0x22,0x22	;41.-60. Pixel, gruen
.db 0x44,0x44,0x44,0x44,0x44,0x44,0x44,0x44,0x44,0x44	;61.-80. Pixel, blau
.db 0x77,0x77,0x77,0x77,0x77,0x77,0x77,0x77,0x77,0x77	;21.-100. Pixel, weiss
.db 0x33,0x33,0x33,0x33,0x33,0x33,0x33,0x33,0x33,0x33	;21.-120. Pixel, gelb
.db 0x55,0x55,0x55,0x55,0x55,0x55,0x55,0x55,0x55,0x55	;21.-140. Pixel, cyan
.db 0x66,0x66,0x66,0x66,0x66,0x66,0x66,0x66,0x66,0x66	;21.-160. Pixel, rotblau
.db 0x70,0x70,0x70,0x70,0x70,0x70,0x70,0x70,0x70,0x70	;21.-180. Pixel, grau
.db  0x00,0x00,0x00,0x00,0x00,0x0F			;181.-192. Pixel, schwarz
;; 80. Zeile:					
.db 0xF0,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00	;erste 10 Bytes = 20 Pixel, schwarz
.db 0x11,0x11,0x11,0x11,0x11,0x11,0x11,0x11,0x11,0x11	;21.-40. Pixel, rot
.db 0x22,0x22,0x22,0x22,0x22,0x22,0x22,0x22,0x22,0x22	;41.-60. Pixel, gruen
.db 0x44,0x44,0x44,0x44,0x44,0x44,0x44,0x44,0x44,0x44	;61.-80. Pixel, blau
.db 0x77,0x77,0x77,0x77,0x77,0x77,0x77,0x77,0x77,0x77	;21.-100. Pixel, weiss
.db 0x33,0x33,0x33,0x33,0x33,0x33,0x33,0x33,0x33,0x33	;21.-120. Pixel, gelb
.db 0x55,0x55,0x55,0x55,0x55,0x55,0x55,0x55,0x55,0x55	;21.-140. Pixel, cyan
.db 0x66,0x66,0x66,0x66,0x66,0x66,0x66,0x66,0x66,0x66	;21.-160. Pixel, rotblau
.db 0x70,0x70,0x70,0x70,0x70,0x70,0x70,0x70,0x70,0x70	;21.-180. Pixel, grau
.db  0x00,0x00,0x00,0x00,0x00,0x0F			;181.-192. Pixel, schwarz
;; 21.-40. Zeile:
.db 0xF0,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00	;erste 10 Bytes = 20 Pixel, schwarz
.db 0x11,0x11,0x11,0x11,0x11,0x11,0x11,0x11,0x11,0x11	;21.-40. Pixel, rot
.db 0x22,0x22,0x22,0x22,0x22,0x22,0x22,0x22,0x22,0x22	;41.-60. Pixel, gruen
.db 0x44,0x44,0x44,0x44,0x44,0x44,0x44,0x44,0x44,0x44	;61.-80. Pixel, blau
.db 0x77,0x77,0x77,0x77,0x77,0x77,0x77,0x77,0x77,0x77	;21.-100. Pixel, weiss
.db 0x33,0x33,0x33,0x33,0x33,0x33,0x33,0x33,0x33,0x33	;21.-120. Pixel, gelb
.db 0x55,0x55,0x55,0x55,0x55,0x55,0x55,0x55,0x55,0x55	;21.-140. Pixel, cyan
.db 0x66,0x66,0x66,0x66,0x66,0x66,0x66,0x66,0x66,0x66	;21.-160. Pixel, rotblau
.db 0x70,0x70,0x70,0x70,0x70,0x70,0x70,0x70,0x70,0x70	;21.-180. Pixel, grau
.db  0x00,0x00,0x00,0x00,0x00,0x0F			;181.-192. Pixel, schwarz
;; 2. Zeile:					
.db 0xF0,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00	;erste 10 Bytes = 20 Pixel, schwarz
.db 0x11,0x11,0x11,0x11,0x11,0x11,0x11,0x11,0x11,0x11	;21.-40. Pixel, rot
.db 0x22,0x22,0x22,0x22,0x22,0x22,0x22,0x22,0x22,0x22	;41.-60. Pixel, gruen
.db 0x44,0x44,0x44,0x44,0x44,0x44,0x44,0x44,0x44,0x44	;61.-80. Pixel, blau
.db 0x77,0x77,0x77,0x77,0x77,0x77,0x77,0x77,0x77,0x77	;21.-100. Pixel, weiss
.db 0x33,0x33,0x33,0x33,0x33,0x33,0x33,0x33,0x33,0x33	;21.-120. Pixel, gelb
.db 0x55,0x55,0x55,0x55,0x55,0x55,0x55,0x55,0x55,0x55	;21.-140. Pixel, cyan
.db 0x66,0x66,0x66,0x66,0x66,0x66,0x66,0x66,0x66,0x66	;21.-160. Pixel, rotblau
.db 0x70,0x70,0x70,0x70,0x70,0x70,0x70,0x70,0x70,0x70	;21.-180. Pixel, grau
.db  0x00,0x00,0x00,0x00,0x00,0x0F			;181.-192. Pixel, schwarz
;; 3. Zeile:					
.db 0xF0,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00	;erste 10 Bytes = 20 Pixel, schwarz
.db 0x11,0x11,0x11,0x11,0x11,0x11,0x11,0x11,0x11,0x11	;21.-40. Pixel, rot
.db 0x22,0x22,0x22,0x22,0x22,0x22,0x22,0x22,0x22,0x22	;41.-60. Pixel, gruen
.db 0x44,0x44,0x44,0x44,0x44,0x44,0x44,0x44,0x44,0x44	;61.-80. Pixel, blau
.db 0x77,0x77,0x77,0x77,0x77,0x77,0x77,0x77,0x77,0x77	;21.-100. Pixel, weiss
.db 0x33,0x33,0x33,0x33,0x33,0x33,0x33,0x33,0x33,0x33	;21.-120. Pixel, gelb
.db 0x55,0x55,0x55,0x55,0x55,0x55,0x55,0x55,0x55,0x55	;21.-140. Pixel, cyan
.db 0x66,0x66,0x66,0x66,0x66,0x66,0x66,0x66,0x66,0x66	;21.-160. Pixel, rotblau
.db 0x70,0x70,0x70,0x70,0x70,0x70,0x70,0x70,0x70,0x70	;21.-180. Pixel, grau
.db  0x00,0x00,0x00,0x00,0x00,0x0F			;181.-192. Pixel, schwarz
;; 4. Zeile:					
.db 0xF0,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00	;erste 10 Bytes = 20 Pixel, schwarz
.db 0x11,0x11,0x11,0x11,0x11,0x11,0x11,0x11,0x11,0x11	;21.-40. Pixel, rot
.db 0x22,0x22,0x22,0x22,0x22,0x22,0x22,0x22,0x22,0x22	;41.-60. Pixel, gruen
.db 0x44,0x44,0x44,0x44,0x44,0x44,0x44,0x44,0x44,0x44	;61.-80. Pixel, blau
.db 0x77,0x77,0x77,0x77,0x77,0x77,0x77,0x77,0x77,0x77	;21.-100. Pixel, weiss
.db 0x33,0x33,0x33,0x33,0x33,0x33,0x33,0x33,0x33,0x33	;21.-120. Pixel, gelb
.db 0x55,0x55,0x55,0x55,0x55,0x55,0x55,0x55,0x55,0x55	;21.-140. Pixel, cyan
.db 0x66,0x66,0x66,0x66,0x66,0x66,0x66,0x66,0x66,0x66	;21.-160. Pixel, rotblau
.db 0x70,0x70,0x70,0x70,0x70,0x70,0x70,0x70,0x70,0x70	;21.-180. Pixel, grau
.db  0x00,0x00,0x00,0x00,0x00,0x0F			;181.-192. Pixel, schwarz
;; 5. Zeile:					
.db 0xF0,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00	;erste 10 Bytes = 20 Pixel, schwarz
.db 0x11,0x11,0x11,0x11,0x11,0x11,0x11,0x11,0x11,0x11	;21.-40. Pixel, rot
.db 0x22,0x22,0x22,0x22,0x22,0x22,0x22,0x22,0x22,0x22	;41.-60. Pixel, gruen
.db 0x44,0x44,0x44,0x44,0x44,0x44,0x44,0x44,0x44,0x44	;61.-80. Pixel, blau
.db 0x77,0x77,0x77,0x77,0x77,0x77,0x77,0x77,0x77,0x77	;21.-100. Pixel, weiss
.db 0x33,0x33,0x33,0x33,0x33,0x33,0x33,0x33,0x33,0x33	;21.-120. Pixel, gelb
.db 0x55,0x55,0x55,0x55,0x55,0x55,0x55,0x55,0x55,0x55	;21.-140. Pixel, cyan
.db 0x66,0x66,0x66,0x66,0x66,0x66,0x66,0x66,0x66,0x66	;21.-160. Pixel, rotblau
.db 0x70,0x70,0x70,0x70,0x70,0x70,0x70,0x70,0x70,0x70	;21.-180. Pixel, grau
.db  0x00,0x00,0x00,0x00,0x00,0x0F			;181.-192. Pixel, schwarz
;; 6. Zeile:					
.db 0xF0,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00	;erste 10 Bytes = 20 Pixel, schwarz
.db 0x11,0x11,0x11,0x11,0x11,0x11,0x11,0x11,0x11,0x11	;21.-40. Pixel, rot
.db 0x22,0x22,0x22,0x22,0x22,0x22,0x22,0x22,0x22,0x22	;41.-60. Pixel, gruen
.db 0x44,0x44,0x44,0x44,0x44,0x44,0x44,0x44,0x44,0x44	;61.-80. Pixel, blau
.db 0x77,0x77,0x77,0x77,0x77,0x77,0x77,0x77,0x77,0x77	;21.-100. Pixel, weiss
.db 0x33,0x33,0x33,0x33,0x33,0x33,0x33,0x33,0x33,0x33	;21.-120. Pixel, gelb
.db 0x55,0x55,0x55,0x55,0x55,0x55,0x55,0x55,0x55,0x55	;21.-140. Pixel, cyan
.db 0x66,0x66,0x66,0x66,0x66,0x66,0x66,0x66,0x66,0x66	;21.-160. Pixel, rotblau
.db 0x70,0x70,0x70,0x70,0x70,0x70,0x70,0x70,0x70,0x70	;21.-180. Pixel, grau
.db  0x00,0x00,0x00,0x00,0x00,0x0F			;181.-192. Pixel, schwarz
;; 7. Zeile:					
.db 0xF0,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00	;erste 10 Bytes = 20 Pixel, schwarz
.db 0x11,0x11,0x11,0x11,0x11,0x11,0x11,0x11,0x11,0x11	;21.-40. Pixel, rot
.db 0x22,0x22,0x22,0x22,0x22,0x22,0x22,0x22,0x22,0x22	;41.-60. Pixel, gruen
.db 0x44,0x44,0x44,0x44,0x44,0x44,0x44,0x44,0x44,0x44	;61.-80. Pixel, blau
.db 0x77,0x77,0x77,0x77,0x77,0x77,0x77,0x77,0x77,0x77	;21.-100. Pixel, weiss
.db 0x33,0x33,0x33,0x33,0x33,0x33,0x33,0x33,0x33,0x33	;21.-120. Pixel, gelb
.db 0x55,0x55,0x55,0x55,0x55,0x55,0x55,0x55,0x55,0x55	;21.-140. Pixel, cyan
.db 0x66,0x66,0x66,0x66,0x66,0x66,0x66,0x66,0x66,0x66	;21.-160. Pixel, rotblau
.db 0x70,0x70,0x70,0x70,0x70,0x70,0x70,0x70,0x70,0x70	;21.-180. Pixel, grau
.db  0x00,0x00,0x00,0x00,0x00,0x0F			;181.-192. Pixel, schwarz
;; 8. Zeile:					
.db 0xF0,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00	;erste 10 Bytes = 20 Pixel, schwarz
.db 0x11,0x11,0x11,0x11,0x11,0x11,0x11,0x11,0x11,0x11	;21.-40. Pixel, rot
.db 0x22,0x22,0x22,0x22,0x22,0x22,0x22,0x22,0x22,0x22	;41.-60. Pixel, gruen
.db 0x44,0x44,0x44,0x44,0x44,0x44,0x44,0x44,0x44,0x44	;61.-80. Pixel, blau
.db 0x77,0x77,0x77,0x77,0x77,0x77,0x77,0x77,0x77,0x77	;21.-100. Pixel, weiss
.db 0x33,0x33,0x33,0x33,0x33,0x33,0x33,0x33,0x33,0x33	;21.-120. Pixel, gelb
.db 0x55,0x55,0x55,0x55,0x55,0x55,0x55,0x55,0x55,0x55	;21.-140. Pixel, cyan
.db 0x66,0x66,0x66,0x66,0x66,0x66,0x66,0x66,0x66,0x66	;21.-160. Pixel, rotblau
.db 0x70,0x70,0x70,0x70,0x70,0x70,0x70,0x70,0x70,0x70	;21.-180. Pixel, grau
.db  0x00,0x00,0x00,0x00,0x00,0x0F			;181.-192. Pixel, schwarz
;; 9. Zeile:					
.db 0xF0,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00	;erste 10 Bytes = 20 Pixel, schwarz
.db 0x11,0x11,0x11,0x11,0x11,0x11,0x11,0x11,0x11,0x11	;21.-40. Pixel, rot
.db 0x22,0x22,0x22,0x22,0x22,0x22,0x22,0x22,0x22,0x22	;41.-60. Pixel, gruen
.db 0x44,0x44,0x44,0x44,0x44,0x44,0x44,0x44,0x44,0x44	;61.-80. Pixel, blau
.db 0x77,0x77,0x77,0x77,0x77,0x77,0x77,0x77,0x77,0x77	;21.-100. Pixel, weiss
.db 0x33,0x33,0x33,0x33,0x33,0x33,0x33,0x33,0x33,0x33	;21.-120. Pixel, gelb
.db 0x55,0x55,0x55,0x55,0x55,0x55,0x55,0x55,0x55,0x55	;21.-140. Pixel, cyan
.db 0x66,0x66,0x66,0x66,0x66,0x66,0x66,0x66,0x66,0x66	;21.-160. Pixel, rotblau
.db 0x70,0x70,0x70,0x70,0x70,0x70,0x70,0x70,0x70,0x70	;21.-180. Pixel, grau
.db  0x00,0x00,0x00,0x00,0x00,0x0F			;181.-192. Pixel, schwarz
;; 90. Zeile:					
.db 0xF0,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00	;erste 10 Bytes = 20 Pixel, schwarz
.db 0x11,0x11,0x11,0x11,0x11,0x11,0x11,0x11,0x11,0x11	;21.-40. Pixel, rot
.db 0x22,0x22,0x22,0x22,0x22,0x22,0x22,0x22,0x22,0x22	;41.-60. Pixel, gruen
.db 0x44,0x44,0x44,0x44,0x44,0x44,0x44,0x44,0x44,0x44	;61.-80. Pixel, blau
.db 0x77,0x77,0x77,0x77,0x77,0x77,0x77,0x77,0x77,0x77	;21.-100. Pixel, weiss
.db 0x33,0x33,0x33,0x33,0x33,0x33,0x33,0x33,0x33,0x33	;21.-120. Pixel, gelb
.db 0x55,0x55,0x55,0x55,0x55,0x55,0x55,0x55,0x55,0x55	;21.-140. Pixel, cyan
.db 0x66,0x66,0x66,0x66,0x66,0x66,0x66,0x66,0x66,0x66	;21.-160. Pixel, rotblau
.db 0x70,0x70,0x70,0x70,0x70,0x70,0x70,0x70,0x70,0x70	;21.-180. Pixel, grau
.db  0x00,0x00,0x00,0x00,0x00,0x0F			;181.-192. Pixel, schwarz
;; 11.-20. Zeile:
.db 0xF0,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00	;erste 10 Bytes = 20 Pixel, schwarz
.db 0x11,0x11,0x11,0x11,0x11,0x11,0x11,0x11,0x11,0x11	;21.-40. Pixel, rot
.db 0x22,0x22,0x22,0x22,0x22,0x22,0x22,0x22,0x22,0x22	;41.-60. Pixel, gruen
.db 0x44,0x44,0x44,0x44,0x44,0x44,0x44,0x44,0x44,0x44	;61.-80. Pixel, blau
.db 0x77,0x77,0x77,0x77,0x77,0x77,0x77,0x77,0x77,0x77	;21.-100. Pixel, weiss
.db 0x33,0x33,0x33,0x33,0x33,0x33,0x33,0x33,0x33,0x33	;21.-120. Pixel, gelb
.db 0x55,0x55,0x55,0x55,0x55,0x55,0x55,0x55,0x55,0x55	;21.-140. Pixel, cyan
.db 0x66,0x66,0x66,0x66,0x66,0x66,0x66,0x66,0x66,0x66	;21.-160. Pixel, rotblau
.db 0x70,0x70,0x70,0x70,0x70,0x70,0x70,0x70,0x70,0x70	;21.-180. Pixel, grau
.db  0x00,0x00,0x00,0x00,0x00,0x0F			;181.-192. Pixel, schwarz
;; 12. Zeile:					
.db 0xF0,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00	;erste 10 Bytes = 20 Pixel, schwarz
.db 0x11,0x11,0x11,0x11,0x11,0x11,0x11,0x11,0x11,0x11	;21.-40. Pixel, rot
.db 0x22,0x22,0x22,0x22,0x22,0x22,0x22,0x22,0x22,0x22	;41.-60. Pixel, gruen
.db 0x44,0x44,0x44,0x44,0x44,0x44,0x44,0x44,0x44,0x44	;61.-80. Pixel, blau
.db 0x77,0x77,0x77,0x77,0x77,0x77,0x77,0x77,0x77,0x77	;21.-100. Pixel, weiss
.db 0x33,0x33,0x33,0x33,0x33,0x33,0x33,0x33,0x33,0x33	;21.-120. Pixel, gelb
.db 0x55,0x55,0x55,0x55,0x55,0x55,0x55,0x55,0x55,0x55	;21.-140. Pixel, cyan
.db 0x66,0x66,0x66,0x66,0x66,0x66,0x66,0x66,0x66,0x66	;21.-160. Pixel, rotblau
.db 0x70,0x70,0x70,0x70,0x70,0x70,0x70,0x70,0x70,0x70	;21.-180. Pixel, grau
.db  0x00,0x00,0x00,0x00,0x00,0x0F			;181.-192. Pixel, schwarz
;; 13. Zeile:					
.db 0xF0,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00	;erste 10 Bytes = 20 Pixel, schwarz
.db 0x11,0x11,0x11,0x11,0x11,0x11,0x11,0x11,0x11,0x11	;21.-40. Pixel, rot
.db 0x22,0x22,0x22,0x22,0x22,0x22,0x22,0x22,0x22,0x22	;41.-60. Pixel, gruen
.db 0x44,0x44,0x44,0x44,0x44,0x44,0x44,0x44,0x44,0x44	;61.-80. Pixel, blau
.db 0x77,0x77,0x77,0x77,0x77,0x77,0x77,0x77,0x77,0x77	;21.-100. Pixel, weiss
.db 0x33,0x33,0x33,0x33,0x33,0x33,0x33,0x33,0x33,0x33	;21.-120. Pixel, gelb
.db 0x55,0x55,0x55,0x55,0x55,0x55,0x55,0x55,0x55,0x55	;21.-140. Pixel, cyan
.db 0x66,0x66,0x66,0x66,0x66,0x66,0x66,0x66,0x66,0x66	;21.-160. Pixel, rotblau
.db 0x70,0x70,0x70,0x70,0x70,0x70,0x70,0x70,0x70,0x70	;21.-180. Pixel, grau
.db  0x00,0x00,0x00,0x00,0x00,0x0F			;181.-192. Pixel, schwarz
;; 14. Zeile:					
.db 0xF0,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00	;erste 10 Bytes = 20 Pixel, schwarz
.db 0x11,0x11,0x11,0x11,0x11,0x11,0x11,0x11,0x11,0x11	;21.-40. Pixel, rot
.db 0x22,0x22,0x22,0x22,0x22,0x22,0x22,0x22,0x22,0x22	;41.-60. Pixel, gruen
.db 0x44,0x44,0x44,0x44,0x44,0x44,0x44,0x44,0x44,0x44	;61.-80. Pixel, blau
.db 0x77,0x77,0x77,0x77,0x77,0x77,0x77,0x77,0x77,0x77	;21.-100. Pixel, weiss
.db 0x33,0x33,0x33,0x33,0x33,0x33,0x33,0x33,0x33,0x33	;21.-120. Pixel, gelb
.db 0x55,0x55,0x55,0x55,0x55,0x55,0x55,0x55,0x55,0x55	;21.-140. Pixel, cyan
.db 0x66,0x66,0x66,0x66,0x66,0x66,0x66,0x66,0x66,0x66	;21.-160. Pixel, rotblau
.db 0x70,0x70,0x70,0x70,0x70,0x70,0x70,0x70,0x70,0x70	;21.-180. Pixel, grau
.db  0x00,0x00,0x00,0x00,0x00,0x0F			;181.-192. Pixel, schwarz
;; 15. Zeile:					
.db 0xF0,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00	;erste 10 Bytes = 20 Pixel, schwarz
.db 0x11,0x11,0x11,0x11,0x11,0x11,0x11,0x11,0x11,0x11	;21.-40. Pixel, rot
.db 0x22,0x22,0x22,0x22,0x22,0x22,0x22,0x22,0x22,0x22	;41.-60. Pixel, gruen
.db 0x44,0x44,0x44,0x44,0x44,0x44,0x44,0x44,0x44,0x44	;61.-80. Pixel, blau
.db 0x77,0x77,0x77,0x77,0x77,0x77,0x77,0x77,0x77,0x77	;21.-100. Pixel, weiss
.db 0x33,0x33,0x33,0x33,0x33,0x33,0x33,0x33,0x33,0x33	;21.-120. Pixel, gelb
.db 0x55,0x55,0x55,0x55,0x55,0x55,0x55,0x55,0x55,0x55	;21.-140. Pixel, cyan
.db 0x66,0x66,0x66,0x66,0x66,0x66,0x66,0x66,0x66,0x66	;21.-160. Pixel, rotblau
.db 0x70,0x70,0x70,0x70,0x70,0x70,0x70,0x70,0x70,0x70	;21.-180. Pixel, grau
.db  0x00,0x00,0x00,0x00,0x00,0x0F			;181.-192. Pixel, schwarz
;; 16. Zeile:					
.db 0xF0,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00	;erste 10 Bytes = 20 Pixel, schwarz
.db 0x11,0x11,0x11,0x11,0x11,0x11,0x11,0x11,0x11,0x11	;21.-40. Pixel, rot
.db 0x22,0x22,0x22,0x22,0x22,0x22,0x22,0x22,0x22,0x22	;41.-60. Pixel, gruen
.db 0x44,0x44,0x44,0x44,0x44,0x44,0x44,0x44,0x44,0x44	;61.-80. Pixel, blau
.db 0x77,0x77,0x77,0x77,0x77,0x77,0x77,0x77,0x77,0x77	;21.-100. Pixel, weiss
.db 0x33,0x33,0x33,0x33,0x33,0x33,0x33,0x33,0x33,0x33	;21.-120. Pixel, gelb
.db 0x55,0x55,0x55,0x55,0x55,0x55,0x55,0x55,0x55,0x55	;21.-140. Pixel, cyan
.db 0x66,0x66,0x66,0x66,0x66,0x66,0x66,0x66,0x66,0x66	;21.-160. Pixel, rotblau
.db 0x70,0x70,0x70,0x70,0x70,0x70,0x70,0x70,0x70,0x70	;21.-180. Pixel, grau
.db  0x00,0x00,0x00,0x00,0x00,0x0F			;181.-192. Pixel, schwarz
;; 17. Zeile:					
.db 0xF0,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00	;erste 10 Bytes = 20 Pixel, schwarz
.db 0x11,0x11,0x11,0x11,0x11,0x11,0x11,0x11,0x11,0x11	;21.-40. Pixel, rot
.db 0x22,0x22,0x22,0x22,0x22,0x22,0x22,0x22,0x22,0x22	;41.-60. Pixel, gruen
.db 0x44,0x44,0x44,0x44,0x44,0x44,0x44,0x44,0x44,0x44	;61.-80. Pixel, blau
.db 0x77,0x77,0x77,0x77,0x77,0x77,0x77,0x77,0x77,0x77	;21.-100. Pixel, weiss
.db 0x33,0x33,0x33,0x33,0x33,0x33,0x33,0x33,0x33,0x33	;21.-120. Pixel, gelb
.db 0x55,0x55,0x55,0x55,0x55,0x55,0x55,0x55,0x55,0x55	;21.-140. Pixel, cyan
.db 0x66,0x66,0x66,0x66,0x66,0x66,0x66,0x66,0x66,0x66	;21.-160. Pixel, rotblau
.db 0x70,0x70,0x70,0x70,0x70,0x70,0x70,0x70,0x70,0x70	;21.-180. Pixel, grau
.db  0x00,0x00,0x00,0x00,0x00,0x0F			;181.-192. Pixel, schwarz
;; 18. Zeile:					
.db 0xF0,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00	;erste 10 Bytes = 20 Pixel, schwarz
.db 0x11,0x11,0x11,0x11,0x11,0x11,0x11,0x11,0x11,0x11	;21.-40. Pixel, rot
.db 0x22,0x22,0x22,0x22,0x22,0x22,0x22,0x22,0x22,0x22	;41.-60. Pixel, gruen
.db 0x44,0x44,0x44,0x44,0x44,0x44,0x44,0x44,0x44,0x44	;61.-80. Pixel, blau
.db 0x77,0x77,0x77,0x77,0x77,0x77,0x77,0x77,0x77,0x77	;21.-100. Pixel, weiss
.db 0x33,0x33,0x33,0x33,0x33,0x33,0x33,0x33,0x33,0x33	;21.-120. Pixel, gelb
.db 0x55,0x55,0x55,0x55,0x55,0x55,0x55,0x55,0x55,0x55	;21.-140. Pixel, cyan
.db 0x66,0x66,0x66,0x66,0x66,0x66,0x66,0x66,0x66,0x66	;21.-160. Pixel, rotblau
.db 0x70,0x70,0x70,0x70,0x70,0x70,0x70,0x70,0x70,0x70	;21.-180. Pixel, grau
.db  0x00,0x00,0x00,0x00,0x00,0x0F			;181.-192. Pixel, schwarz
;; 19. Zeile:					
.db 0xF0,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00	;erste 10 Bytes = 20 Pixel, schwarz
.db 0x11,0x11,0x11,0x11,0x11,0x11,0x11,0x11,0x11,0x11	;21.-40. Pixel, rot
.db 0x22,0x22,0x22,0x22,0x22,0x22,0x22,0x22,0x22,0x22	;41.-60. Pixel, gruen
.db 0x44,0x44,0x44,0x44,0x44,0x44,0x44,0x44,0x44,0x44	;61.-80. Pixel, blau
.db 0x77,0x77,0x77,0x77,0x77,0x77,0x77,0x77,0x77,0x77	;21.-100. Pixel, weiss
.db 0x33,0x33,0x33,0x33,0x33,0x33,0x33,0x33,0x33,0x33	;21.-120. Pixel, gelb
.db 0x55,0x55,0x55,0x55,0x55,0x55,0x55,0x55,0x55,0x55	;21.-140. Pixel, cyan
.db 0x66,0x66,0x66,0x66,0x66,0x66,0x66,0x66,0x66,0x66	;21.-160. Pixel, rotblau
.db 0x70,0x70,0x70,0x70,0x70,0x70,0x70,0x70,0x70,0x70	;21.-180. Pixel, grau
.db  0x00,0x00,0x00,0x00,0x00,0x0F			;181.-192. Pixel, schwarz
;; 100.
;; 80. Zeile:					
.db 0xF0,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00	;erste 10 Bytes = 20 Pixel, schwarz
.db 0x11,0x11,0x11,0x11,0x11,0x11,0x11,0x11,0x11,0x11	;21.-40. Pixel, rot
.db 0x22,0x22,0x22,0x22,0x22,0x22,0x22,0x22,0x22,0x22	;41.-60. Pixel, gruen
.db 0x44,0x44,0x44,0x44,0x44,0x44,0x44,0x44,0x44,0x44	;61.-80. Pixel, blau
.db 0x77,0x77,0x77,0x77,0x77,0x77,0x77,0x77,0x77,0x77	;21.-100. Pixel, weiss
.db 0x33,0x33,0x33,0x33,0x33,0x33,0x33,0x33,0x33,0x33	;21.-120. Pixel, gelb
.db 0x55,0x55,0x55,0x55,0x55,0x55,0x55,0x55,0x55,0x55	;21.-140. Pixel, cyan
.db 0x66,0x66,0x66,0x66,0x66,0x66,0x66,0x66,0x66,0x66	;21.-160. Pixel, rotblau
.db 0x70,0x70,0x70,0x70,0x70,0x70,0x70,0x70,0x70,0x70	;21.-180. Pixel, grau
.db  0x00,0x00,0x00,0x00,0x00,0x0F			;181.-192. Pixel, schwarz
;; 21.-40. Zeile:
.db 0xF0,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00	;erste 10 Bytes = 20 Pixel, schwarz
.db 0x11,0x11,0x11,0x11,0x11,0x11,0x11,0x11,0x11,0x11	;21.-40. Pixel, rot
.db 0x22,0x22,0x22,0x22,0x22,0x22,0x22,0x22,0x22,0x22	;41.-60. Pixel, gruen
.db 0x44,0x44,0x44,0x44,0x44,0x44,0x44,0x44,0x44,0x44	;61.-80. Pixel, blau
.db 0x77,0x77,0x77,0x77,0x77,0x77,0x77,0x77,0x77,0x77	;21.-100. Pixel, weiss
.db 0x33,0x33,0x33,0x33,0x33,0x33,0x33,0x33,0x33,0x33	;21.-120. Pixel, gelb
.db 0x55,0x55,0x55,0x55,0x55,0x55,0x55,0x55,0x55,0x55	;21.-140. Pixel, cyan
.db 0x66,0x66,0x66,0x66,0x66,0x66,0x66,0x66,0x66,0x66	;21.-160. Pixel, rotblau
.db 0x70,0x70,0x70,0x70,0x70,0x70,0x70,0x70,0x70,0x70	;21.-180. Pixel, grau
.db  0x00,0x00,0x00,0x00,0x00,0x0F			;181.-192. Pixel, schwarz
;; 2. Zeile:					
.db 0xF0,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00	;erste 10 Bytes = 20 Pixel, schwarz
.db 0x11,0x11,0x11,0x11,0x11,0x11,0x11,0x11,0x11,0x11	;21.-40. Pixel, rot
.db 0x22,0x22,0x22,0x22,0x22,0x22,0x22,0x22,0x22,0x22	;41.-60. Pixel, gruen
.db 0x44,0x44,0x44,0x44,0x44,0x44,0x44,0x44,0x44,0x44	;61.-80. Pixel, blau
.db 0x77,0x77,0x77,0x77,0x77,0x77,0x77,0x77,0x77,0x77	;21.-100. Pixel, weiss
.db 0x33,0x33,0x33,0x33,0x33,0x33,0x33,0x33,0x33,0x33	;21.-120. Pixel, gelb
.db 0x55,0x55,0x55,0x55,0x55,0x55,0x55,0x55,0x55,0x55	;21.-140. Pixel, cyan
.db 0x66,0x66,0x66,0x66,0x66,0x66,0x66,0x66,0x66,0x66	;21.-160. Pixel, rotblau
.db 0x70,0x70,0x70,0x70,0x70,0x70,0x70,0x70,0x70,0x70	;21.-180. Pixel, grau
.db  0x00,0x00,0x00,0x00,0x00,0x0F			;181.-192. Pixel, schwarz
;; 3. Zeile:					
.db 0xF0,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00	;erste 10 Bytes = 20 Pixel, schwarz
.db 0x11,0x11,0x11,0x11,0x11,0x11,0x11,0x11,0x11,0x11	;21.-40. Pixel, rot
.db 0x22,0x22,0x22,0x22,0x22,0x22,0x22,0x22,0x22,0x22	;41.-60. Pixel, gruen
.db 0x44,0x44,0x44,0x44,0x44,0x44,0x44,0x44,0x44,0x44	;61.-80. Pixel, blau
.db 0x77,0x77,0x77,0x77,0x77,0x77,0x77,0x77,0x77,0x77	;21.-100. Pixel, weiss
.db 0x33,0x33,0x33,0x33,0x33,0x33,0x33,0x33,0x33,0x33	;21.-120. Pixel, gelb
.db 0x55,0x55,0x55,0x55,0x55,0x55,0x55,0x55,0x55,0x55	;21.-140. Pixel, cyan
.db 0x66,0x66,0x66,0x66,0x66,0x66,0x66,0x66,0x66,0x66	;21.-160. Pixel, rotblau
.db 0x70,0x70,0x70,0x70,0x70,0x70,0x70,0x70,0x70,0x70	;21.-180. Pixel, grau
.db  0x00,0x00,0x00,0x00,0x00,0x0F			;181.-192. Pixel, schwarz
;; 4. Zeile:					
.db 0xF0,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00	;erste 10 Bytes = 20 Pixel, schwarz
.db 0x11,0x11,0x11,0x11,0x11,0x11,0x11,0x11,0x11,0x11	;21.-40. Pixel, rot
.db 0x22,0x22,0x22,0x22,0x22,0x22,0x22,0x22,0x22,0x22	;41.-60. Pixel, gruen
.db 0x44,0x44,0x44,0x44,0x44,0x44,0x44,0x44,0x44,0x44	;61.-80. Pixel, blau
.db 0x77,0x77,0x77,0x77,0x77,0x77,0x77,0x77,0x77,0x77	;21.-100. Pixel, weiss
.db 0x33,0x33,0x33,0x33,0x33,0x33,0x33,0x33,0x33,0x33	;21.-120. Pixel, gelb
.db 0x55,0x55,0x55,0x55,0x55,0x55,0x55,0x55,0x55,0x55	;21.-140. Pixel, cyan
.db 0x66,0x66,0x66,0x66,0x66,0x66,0x66,0x66,0x66,0x66	;21.-160. Pixel, rotblau
.db 0x70,0x70,0x70,0x70,0x70,0x70,0x70,0x70,0x70,0x70	;21.-180. Pixel, grau
.db  0x00,0x00,0x00,0x00,0x00,0x0F			;181.-192. Pixel, schwarz
;; 5. Zeile:					
.db 0xF0,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00	;erste 10 Bytes = 20 Pixel, schwarz
.db 0x11,0x11,0x11,0x11,0x11,0x11,0x11,0x11,0x11,0x11	;21.-40. Pixel, rot
.db 0x22,0x22,0x22,0x22,0x22,0x22,0x22,0x22,0x22,0x22	;41.-60. Pixel, gruen
.db 0x44,0x44,0x44,0x44,0x44,0x44,0x44,0x44,0x44,0x44	;61.-80. Pixel, blau
.db 0x77,0x77,0x77,0x77,0x77,0x77,0x77,0x77,0x77,0x77	;21.-100. Pixel, weiss
.db 0x33,0x33,0x33,0x33,0x33,0x33,0x33,0x33,0x33,0x33	;21.-120. Pixel, gelb
.db 0x55,0x55,0x55,0x55,0x55,0x55,0x55,0x55,0x55,0x55	;21.-140. Pixel, cyan
.db 0x66,0x66,0x66,0x66,0x66,0x66,0x66,0x66,0x66,0x66	;21.-160. Pixel, rotblau
.db 0x70,0x70,0x70,0x70,0x70,0x70,0x70,0x70,0x70,0x70	;21.-180. Pixel, grau
.db  0x00,0x00,0x00,0x00,0x00,0x0F			;181.-192. Pixel, schwarz
;; 6. Zeile:					
.db 0xF0,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00	;erste 10 Bytes = 20 Pixel, schwarz
.db 0x11,0x11,0x11,0x11,0x11,0x11,0x11,0x11,0x11,0x11	;21.-40. Pixel, rot
.db 0x22,0x22,0x22,0x22,0x22,0x22,0x22,0x22,0x22,0x22	;41.-60. Pixel, gruen
.db 0x44,0x44,0x44,0x44,0x44,0x44,0x44,0x44,0x44,0x44	;61.-80. Pixel, blau
.db 0x77,0x77,0x77,0x77,0x77,0x77,0x77,0x77,0x77,0x77	;21.-100. Pixel, weiss
.db 0x33,0x33,0x33,0x33,0x33,0x33,0x33,0x33,0x33,0x33	;21.-120. Pixel, gelb
.db 0x55,0x55,0x55,0x55,0x55,0x55,0x55,0x55,0x55,0x55	;21.-140. Pixel, cyan
.db 0x66,0x66,0x66,0x66,0x66,0x66,0x66,0x66,0x66,0x66	;21.-160. Pixel, rotblau
.db 0x70,0x70,0x70,0x70,0x70,0x70,0x70,0x70,0x70,0x70	;21.-180. Pixel, grau
.db  0x00,0x00,0x00,0x00,0x00,0x0F			;181.-192. Pixel, schwarz
;; 7. Zeile:					
.db 0xF0,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00	;erste 10 Bytes = 20 Pixel, schwarz
.db 0x11,0x11,0x11,0x11,0x11,0x11,0x11,0x11,0x11,0x11	;21.-40. Pixel, rot
.db 0x22,0x22,0x22,0x22,0x22,0x22,0x22,0x22,0x22,0x22	;41.-60. Pixel, gruen
.db 0x44,0x44,0x44,0x44,0x44,0x44,0x44,0x44,0x44,0x44	;61.-80. Pixel, blau
.db 0x77,0x77,0x77,0x77,0x77,0x77,0x77,0x77,0x77,0x77	;21.-100. Pixel, weiss
.db 0x33,0x33,0x33,0x33,0x33,0x33,0x33,0x33,0x33,0x33	;21.-120. Pixel, gelb
.db 0x55,0x55,0x55,0x55,0x55,0x55,0x55,0x55,0x55,0x55	;21.-140. Pixel, cyan
.db 0x66,0x66,0x66,0x66,0x66,0x66,0x66,0x66,0x66,0x66	;21.-160. Pixel, rotblau
.db 0x70,0x70,0x70,0x70,0x70,0x70,0x70,0x70,0x70,0x70	;21.-180. Pixel, grau
.db  0x00,0x00,0x00,0x00,0x00,0x0F			;181.-192. Pixel, schwarz
;; 8. Zeile:					
.db 0xF0,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00	;erste 10 Bytes = 20 Pixel, schwarz
.db 0x11,0x11,0x11,0x11,0x11,0x11,0x11,0x11,0x11,0x11	;21.-40. Pixel, rot
.db 0x22,0x22,0x22,0x22,0x22,0x22,0x22,0x22,0x22,0x22	;41.-60. Pixel, gruen
.db 0x44,0x44,0x44,0x44,0x44,0x44,0x44,0x44,0x44,0x44	;61.-80. Pixel, blau
.db 0x77,0x77,0x77,0x77,0x77,0x77,0x77,0x77,0x77,0x77	;21.-100. Pixel, weiss
.db 0x33,0x33,0x33,0x33,0x33,0x33,0x33,0x33,0x33,0x33	;21.-120. Pixel, gelb
.db 0x55,0x55,0x55,0x55,0x55,0x55,0x55,0x55,0x55,0x55	;21.-140. Pixel, cyan
.db 0x66,0x66,0x66,0x66,0x66,0x66,0x66,0x66,0x66,0x66	;21.-160. Pixel, rotblau
.db 0x70,0x70,0x70,0x70,0x70,0x70,0x70,0x70,0x70,0x70	;21.-180. Pixel, grau
.db  0x00,0x00,0x00,0x00,0x00,0x0F			;181.-192. Pixel, schwarz
;; 9. Zeile:					
.db 0xF0,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00	;erste 10 Bytes = 20 Pixel, schwarz
.db 0x11,0x11,0x11,0x11,0x11,0x11,0x11,0x11,0x11,0x11	;21.-40. Pixel, rot
.db 0x22,0x22,0x22,0x22,0x22,0x22,0x22,0x22,0x22,0x22	;41.-60. Pixel, gruen
.db 0x44,0x44,0x44,0x44,0x44,0x44,0x44,0x44,0x44,0x44	;61.-80. Pixel, blau
.db 0x77,0x77,0x77,0x77,0x77,0x77,0x77,0x77,0x77,0x77	;21.-100. Pixel, weiss
.db 0x33,0x33,0x33,0x33,0x33,0x33,0x33,0x33,0x33,0x33	;21.-120. Pixel, gelb
.db 0x55,0x55,0x55,0x55,0x55,0x55,0x55,0x55,0x55,0x55	;21.-140. Pixel, cyan
.db 0x66,0x66,0x66,0x66,0x66,0x66,0x66,0x66,0x66,0x66	;21.-160. Pixel, rotblau
.db 0x70,0x70,0x70,0x70,0x70,0x70,0x70,0x70,0x70,0x70	;21.-180. Pixel, grau
.db  0x00,0x00,0x00,0x00,0x00,0x0F			;181.-192. Pixel, schwarz
;; 90. Zeile:					
.db 0xF0,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00	;erste 10 Bytes = 20 Pixel, schwarz
.db 0x11,0x11,0x11,0x11,0x11,0x11,0x11,0x11,0x11,0x11	;21.-40. Pixel, rot
.db 0x22,0x22,0x22,0x22,0x22,0x22,0x22,0x22,0x22,0x22	;41.-60. Pixel, gruen
.db 0x44,0x44,0x44,0x44,0x44,0x44,0x44,0x44,0x44,0x44	;61.-80. Pixel, blau
.db 0x77,0x77,0x77,0x77,0x77,0x77,0x77,0x77,0x77,0x77	;21.-100. Pixel, weiss
.db 0x33,0x33,0x33,0x33,0x33,0x33,0x33,0x33,0x33,0x33	;21.-120. Pixel, gelb
.db 0x55,0x55,0x55,0x55,0x55,0x55,0x55,0x55,0x55,0x55	;21.-140. Pixel, cyan
.db 0x66,0x66,0x66,0x66,0x66,0x66,0x66,0x66,0x66,0x66	;21.-160. Pixel, rotblau
.db 0x70,0x70,0x70,0x70,0x70,0x70,0x70,0x70,0x70,0x70	;21.-180. Pixel, grau
.db  0x00,0x00,0x00,0x00,0x00,0x0F			;181.-192. Pixel, schwarz
;; 11.-20. Zeile:
.db 0xF0,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00	;erste 10 Bytes = 20 Pixel, schwarz
.db 0x11,0x11,0x11,0x11,0x11,0x11,0x11,0x11,0x11,0x11	;21.-40. Pixel, rot
.db 0x22,0x22,0x22,0x22,0x22,0x22,0x22,0x22,0x22,0x22	;41.-60. Pixel, gruen
.db 0x44,0x44,0x44,0x44,0x44,0x44,0x44,0x44,0x44,0x44	;61.-80. Pixel, blau
.db 0x77,0x77,0x77,0x77,0x77,0x77,0x77,0x77,0x77,0x77	;21.-100. Pixel, weiss
.db 0x33,0x33,0x33,0x33,0x33,0x33,0x33,0x33,0x33,0x33	;21.-120. Pixel, gelb
.db 0x55,0x55,0x55,0x55,0x55,0x55,0x55,0x55,0x55,0x55	;21.-140. Pixel, cyan
.db 0x66,0x66,0x66,0x66,0x66,0x66,0x66,0x66,0x66,0x66	;21.-160. Pixel, rotblau
.db 0x70,0x70,0x70,0x70,0x70,0x70,0x70,0x70,0x70,0x70	;21.-180. Pixel, grau
.db  0x00,0x00,0x00,0x00,0x00,0x0F			;181.-192. Pixel, schwarz
;; 12. Zeile:					
.db 0xF0,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00	;erste 10 Bytes = 20 Pixel, schwarz
.db 0x11,0x11,0x11,0x11,0x11,0x11,0x11,0x11,0x11,0x11	;21.-40. Pixel, rot
.db 0x22,0x22,0x22,0x22,0x22,0x22,0x22,0x22,0x22,0x22	;41.-60. Pixel, gruen
.db 0x44,0x44,0x44,0x44,0x44,0x44,0x44,0x44,0x44,0x44	;61.-80. Pixel, blau
.db 0x77,0x77,0x77,0x77,0x77,0x77,0x77,0x77,0x77,0x77	;21.-100. Pixel, weiss
.db 0x33,0x33,0x33,0x33,0x33,0x33,0x33,0x33,0x33,0x33	;21.-120. Pixel, gelb
.db 0x55,0x55,0x55,0x55,0x55,0x55,0x55,0x55,0x55,0x55	;21.-140. Pixel, cyan
.db 0x66,0x66,0x66,0x66,0x66,0x66,0x66,0x66,0x66,0x66	;21.-160. Pixel, rotblau
.db 0x70,0x70,0x70,0x70,0x70,0x70,0x70,0x70,0x70,0x70	;21.-180. Pixel, grau
.db  0x00,0x00,0x00,0x00,0x00,0x0F			;181.-192. Pixel, schwarz
;; 13. Zeile:					
.db 0xF0,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00	;erste 10 Bytes = 20 Pixel, schwarz
.db 0x11,0x11,0x11,0x11,0x11,0x11,0x11,0x11,0x11,0x11	;21.-40. Pixel, rot
.db 0x22,0x22,0x22,0x22,0x22,0x22,0x22,0x22,0x22,0x22	;41.-60. Pixel, gruen
.db 0x44,0x44,0x44,0x44,0x44,0x44,0x44,0x44,0x44,0x44	;61.-80. Pixel, blau
.db 0x77,0x77,0x77,0x77,0x77,0x77,0x77,0x77,0x77,0x77	;21.-100. Pixel, weiss
.db 0x33,0x33,0x33,0x33,0x33,0x33,0x33,0x33,0x33,0x33	;21.-120. Pixel, gelb
.db 0x55,0x55,0x55,0x55,0x55,0x55,0x55,0x55,0x55,0x55	;21.-140. Pixel, cyan
.db 0x66,0x66,0x66,0x66,0x66,0x66,0x66,0x66,0x66,0x66	;21.-160. Pixel, rotblau
.db 0x70,0x70,0x70,0x70,0x70,0x70,0x70,0x70,0x70,0x70	;21.-180. Pixel, grau
.db  0x00,0x00,0x00,0x00,0x00,0x0F			;181.-192. Pixel, schwarz
;; 14. Zeile:					
.db 0xF0,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00	;erste 10 Bytes = 20 Pixel, schwarz
.db 0x11,0x11,0x11,0x11,0x11,0x11,0x11,0x11,0x11,0x11	;21.-40. Pixel, rot
.db 0x22,0x22,0x22,0x22,0x22,0x22,0x22,0x22,0x22,0x22	;41.-60. Pixel, gruen
.db 0x44,0x44,0x44,0x44,0x44,0x44,0x44,0x44,0x44,0x44	;61.-80. Pixel, blau
.db 0x77,0x77,0x77,0x77,0x77,0x77,0x77,0x77,0x77,0x77	;21.-100. Pixel, weiss
.db 0x33,0x33,0x33,0x33,0x33,0x33,0x33,0x33,0x33,0x33	;21.-120. Pixel, gelb
.db 0x55,0x55,0x55,0x55,0x55,0x55,0x55,0x55,0x55,0x55	;21.-140. Pixel, cyan
.db 0x66,0x66,0x66,0x66,0x66,0x66,0x66,0x66,0x66,0x66	;21.-160. Pixel, rotblau
.db 0x70,0x70,0x70,0x70,0x70,0x70,0x70,0x70,0x70,0x70	;21.-180. Pixel, grau
.db  0x00,0x00,0x00,0x00,0x00,0x0F			;181.-192. Pixel, schwarz
;; 15. Zeile:					
.db 0xF0,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00	;erste 10 Bytes = 20 Pixel, schwarz
.db 0x11,0x11,0x11,0x11,0x11,0x11,0x11,0x11,0x11,0x11	;21.-40. Pixel, rot
.db 0x22,0x22,0x22,0x22,0x22,0x22,0x22,0x22,0x22,0x22	;41.-60. Pixel, gruen
.db 0x44,0x44,0x44,0x44,0x44,0x44,0x44,0x44,0x44,0x44	;61.-80. Pixel, blau
.db 0x77,0x77,0x77,0x77,0x77,0x77,0x77,0x77,0x77,0x77	;21.-100. Pixel, weiss
.db 0x33,0x33,0x33,0x33,0x33,0x33,0x33,0x33,0x33,0x33	;21.-120. Pixel, gelb
.db 0x55,0x55,0x55,0x55,0x55,0x55,0x55,0x55,0x55,0x55	;21.-140. Pixel, cyan
.db 0x66,0x66,0x66,0x66,0x66,0x66,0x66,0x66,0x66,0x66	;21.-160. Pixel, rotblau
.db 0x70,0x70,0x70,0x70,0x70,0x70,0x70,0x70,0x70,0x70	;21.-180. Pixel, grau
.db  0x00,0x00,0x00,0x00,0x00,0x0F			;181.-192. Pixel, schwarz
;; 16. Zeile:					
.db 0xF0,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00	;erste 10 Bytes = 20 Pixel, schwarz
.db 0x11,0x11,0x11,0x11,0x11,0x11,0x11,0x11,0x11,0x11	;21.-40. Pixel, rot
.db 0x22,0x22,0x22,0x22,0x22,0x22,0x22,0x22,0x22,0x22	;41.-60. Pixel, gruen
.db 0x44,0x44,0x44,0x44,0x44,0x44,0x44,0x44,0x44,0x44	;61.-80. Pixel, blau
.db 0x77,0x77,0x77,0x77,0x77,0x77,0x77,0x77,0x77,0x77	;21.-100. Pixel, weiss
.db 0x33,0x33,0x33,0x33,0x33,0x33,0x33,0x33,0x33,0x33	;21.-120. Pixel, gelb
.db 0x55,0x55,0x55,0x55,0x55,0x55,0x55,0x55,0x55,0x55	;21.-140. Pixel, cyan
.db 0x66,0x66,0x66,0x66,0x66,0x66,0x66,0x66,0x66,0x66	;21.-160. Pixel, rotblau
.db 0x70,0x70,0x70,0x70,0x70,0x70,0x70,0x70,0x70,0x70	;21.-180. Pixel, grau
.db  0x00,0x00,0x00,0x00,0x00,0x0F			;181.-192. Pixel, schwarz
;; 17. Zeile:					
.db 0xF0,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00	;erste 10 Bytes = 20 Pixel, schwarz
.db 0x11,0x11,0x11,0x11,0x11,0x11,0x11,0x11,0x11,0x11	;21.-40. Pixel, rot
.db 0x22,0x22,0x22,0x22,0x22,0x22,0x22,0x22,0x22,0x22	;41.-60. Pixel, gruen
.db 0x44,0x44,0x44,0x44,0x44,0x44,0x44,0x44,0x44,0x44	;61.-80. Pixel, blau
.db 0x77,0x77,0x77,0x77,0x77,0x77,0x77,0x77,0x77,0x77	;21.-100. Pixel, weiss
.db 0x33,0x33,0x33,0x33,0x33,0x33,0x33,0x33,0x33,0x33	;21.-120. Pixel, gelb
.db 0x55,0x55,0x55,0x55,0x55,0x55,0x55,0x55,0x55,0x55	;21.-140. Pixel, cyan
.db 0x66,0x66,0x66,0x66,0x66,0x66,0x66,0x66,0x66,0x66	;21.-160. Pixel, rotblau
.db 0x70,0x70,0x70,0x70,0x70,0x70,0x70,0x70,0x70,0x70	;21.-180. Pixel, grau
.db  0x00,0x00,0x00,0x00,0x00,0x0F			;181.-192. Pixel, schwarz
;; 18. Zeile:					
.db 0xF0,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00	;erste 10 Bytes = 20 Pixel, schwarz
.db 0x11,0x11,0x11,0x11,0x11,0x11,0x11,0x11,0x11,0x11	;21.-40. Pixel, rot
.db 0x22,0x22,0x22,0x22,0x22,0x22,0x22,0x22,0x22,0x22	;41.-60. Pixel, gruen
.db 0x44,0x44,0x44,0x44,0x44,0x44,0x44,0x44,0x44,0x44	;61.-80. Pixel, blau
.db 0x77,0x77,0x77,0x77,0x77,0x77,0x77,0x77,0x77,0x77	;21.-100. Pixel, weiss
.db 0x33,0x33,0x33,0x33,0x33,0x33,0x33,0x33,0x33,0x33	;21.-120. Pixel, gelb
.db 0x55,0x55,0x55,0x55,0x55,0x55,0x55,0x55,0x55,0x55	;21.-140. Pixel, cyan
.db 0x66,0x66,0x66,0x66,0x66,0x66,0x66,0x66,0x66,0x66	;21.-160. Pixel, rotblau
.db 0x70,0x70,0x70,0x70,0x70,0x70,0x70,0x70,0x70,0x70	;21.-180. Pixel, grau
.db  0x00,0x00,0x00,0x00,0x00,0x0F			;181.-192. Pixel, schwarz
;; 19. Zeile:					
.db 0xF0,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00	;erste 10 Bytes = 20 Pixel, schwarz
.db 0x11,0x11,0x11,0x11,0x11,0x11,0x11,0x11,0x11,0x11	;21.-40. Pixel, rot
.db 0x22,0x22,0x22,0x22,0x22,0x22,0x22,0x22,0x22,0x22	;41.-60. Pixel, gruen
.db 0x44,0x44,0x44,0x44,0x44,0x44,0x44,0x44,0x44,0x44	;61.-80. Pixel, blau
.db 0x77,0x77,0x77,0x77,0x77,0x77,0x77,0x77,0x77,0x77	;21.-100. Pixel, weiss
.db 0x33,0x33,0x33,0x33,0x33,0x33,0x33,0x33,0x33,0x33	;21.-120. Pixel, gelb
.db 0x55,0x55,0x55,0x55,0x55,0x55,0x55,0x55,0x55,0x55	;21.-140. Pixel, cyan
.db 0x66,0x66,0x66,0x66,0x66,0x66,0x66,0x66,0x66,0x66	;21.-160. Pixel, rotblau
.db 0x70,0x70,0x70,0x70,0x70,0x70,0x70,0x70,0x70,0x70	;21.-180. Pixel, grau
.db  0x00,0x00,0x00,0x00,0x00,0x0F			;181.-192. Pixel, schwarz
;; 120. Zeile:					
.db 0xF0,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00	;erste 10 Bytes = 20 Pixel, schwarz
.db 0x11,0x11,0x11,0x11,0x11,0x11,0x11,0x11,0x11,0x11	;21.-40. Pixel, rot
.db 0x22,0x22,0x22,0x22,0x22,0x22,0x22,0x22,0x22,0x22	;41.-60. Pixel, gruen
.db 0x44,0x44,0x44,0x44,0x44,0x44,0x44,0x44,0x44,0x44	;61.-80. Pixel, blau
.db 0x77,0x77,0x77,0x77,0x77,0x77,0x77,0x77,0x77,0x77	;21.-100. Pixel, weiss
.db 0x33,0x33,0x33,0x33,0x33,0x33,0x33,0x33,0x33,0x33	;21.-120. Pixel, gelb
.db 0x55,0x55,0x55,0x55,0x55,0x55,0x55,0x55,0x55,0x55	;21.-140. Pixel, cyan
.db 0x66,0x66,0x66,0x66,0x66,0x66,0x66,0x66,0x66,0x66	;21.-160. Pixel, rotblau
.db 0x70,0x70,0x70,0x70,0x70,0x70,0x70,0x70,0x70,0x70	;21.-180. Pixel, grau
.db  0x00,0x00,0x00,0x00,0x00,0x0F			;181.-192. Pixel, schwarz
;; 121. Zeile:
.db 0xF0,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00	;erste 10 Bytes = 20 Pixel, schwarz
.db 0x11,0x11,0x11,0x11,0x11,0x11,0x11,0x11,0x11,0x11	;21.-40. Pixel, rot
.db 0x22,0x22,0x22,0x22,0x22,0x22,0x22,0x22,0x22,0x22	;41.-60. Pixel, gruen
.db 0x44,0x44,0x44,0x44,0x44,0x44,0x44,0x44,0x44,0x44	;61.-80. Pixel, blau
.db 0x77,0x77,0x77,0x77,0x77,0x77,0x77,0x77,0x77,0x77	;21.-100. Pixel, weiss
.db 0x33,0x33,0x33,0x33,0x33,0x33,0x33,0x33,0x33,0x33	;21.-120. Pixel, gelb
.db 0x55,0x55,0x55,0x55,0x55,0x55,0x55,0x55,0x55,0x55	;21.-140. Pixel, cyan
.db 0x66,0x66,0x66,0x66,0x66,0x66,0x66,0x66,0x66,0x66	;21.-160. Pixel, rotblau
.db 0x70,0x70,0x70,0x70,0x70,0x70,0x70,0x70,0x70,0x70	;21.-180. Pixel, grau
.db  0x00,0x00,0x00,0x00,0x00,0x0F			;181.-192. Pixel, schwarz
;; 122. Zeile:					
.db 0xF0,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00	;erste 10 Bytes = 20 Pixel, schwarz
.db 0x11,0x11,0x11,0x11,0x11,0x11,0x11,0x11,0x11,0x11	;21.-40. Pixel, rot
.db 0x22,0x22,0x22,0x22,0x22,0x22,0x22,0x22,0x22,0x22	;41.-60. Pixel, gruen
.db 0x44,0x44,0x44,0x44,0x44,0x44,0x44,0x44,0x44,0x44	;61.-80. Pixel, blau
.db 0x77,0x77,0x77,0x77,0x77,0x77,0x77,0x77,0x77,0x77	;21.-100. Pixel, weiss
.db 0x33,0x33,0x33,0x33,0x33,0x33,0x33,0x33,0x33,0x33	;21.-120. Pixel, gelb
.db 0x55,0x55,0x55,0x55,0x55,0x55,0x55,0x55,0x55,0x55	;21.-140. Pixel, cyan
.db 0x66,0x66,0x66,0x66,0x66,0x66,0x66,0x66,0x66,0x66	;21.-160. Pixel, rotblau
.db 0x70,0x70,0x70,0x70,0x70,0x70,0x70,0x70,0x70,0x70	;21.-180. Pixel, grau
.db  0x00,0x00,0x00,0x00,0x00,0x0F			;181.-192. Pixel, schwarz
;; 123. Zeile:					
.db 0xF0,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00	;erste 10 Bytes = 20 Pixel, schwarz
.db 0x11,0x11,0x11,0x11,0x11,0x11,0x11,0x11,0x11,0x11	;21.-40. Pixel, rot
.db 0x22,0x22,0x22,0x22,0x22,0x22,0x22,0x22,0x22,0x22	;41.-60. Pixel, gruen
.db 0x44,0x44,0x44,0x44,0x44,0x44,0x44,0x44,0x44,0x44	;61.-80. Pixel, blau
.db 0x77,0x77,0x77,0x77,0x77,0x77,0x77,0x77,0x77,0x77	;21.-100. Pixel, weiss
.db 0x33,0x33,0x33,0x33,0x33,0x33,0x33,0x33,0x33,0x33	;21.-120. Pixel, gelb
.db 0x55,0x55,0x55,0x55,0x55,0x55,0x55,0x55,0x55,0x55	;21.-140. Pixel, cyan
.db 0x66,0x66,0x66,0x66,0x66,0x66,0x66,0x66,0x66,0x66	;21.-160. Pixel, rotblau
.db 0x70,0x70,0x70,0x70,0x70,0x70,0x70,0x70,0x70,0x70	;21.-180. Pixel, grau
.db  0x00,0x00,0x00,0x00,0x00,0x0F			;181.-192. Pixel, schwarz
;; 124. Zeile:					
.db 0xF0,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00	;erste 10 Bytes = 20 Pixel, schwarz
.db 0x11,0x11,0x11,0x11,0x11,0x11,0x11,0x11,0x11,0x11	;21.-40. Pixel, rot
.db 0x22,0x22,0x22,0x22,0x22,0x22,0x22,0x22,0x22,0x22	;41.-60. Pixel, gruen
.db 0x44,0x44,0x44,0x44,0x44,0x44,0x44,0x44,0x44,0x44	;61.-80. Pixel, blau
.db 0x77,0x77,0x77,0x77,0x77,0x77,0x77,0x77,0x77,0x77	;21.-100. Pixel, weiss
.db 0x33,0x33,0x33,0x33,0x33,0x33,0x33,0x33,0x33,0x33	;21.-120. Pixel, gelb
.db 0x55,0x55,0x55,0x55,0x55,0x55,0x55,0x55,0x55,0x55	;21.-140. Pixel, cyan
.db 0x66,0x66,0x66,0x66,0x66,0x66,0x66,0x66,0x66,0x66	;21.-160. Pixel, rotblau
.db 0x70,0x70,0x70,0x70,0x70,0x70,0x70,0x70,0x70,0x70	;21.-180. Pixel, grau
.db  0x00,0x00,0x00,0x00,0x00,0x0F			;181.-192. Pixel, schwarz
;; 125. Zeile:					
.db 0xF0,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00	;erste 10 Bytes = 20 Pixel, schwarz
.db 0x11,0x11,0x11,0x11,0x11,0x11,0x11,0x11,0x11,0x11	;21.-40. Pixel, rot
.db 0x22,0x22,0x22,0x22,0x22,0x22,0x22,0x22,0x22,0x22	;41.-60. Pixel, gruen
.db 0x44,0x44,0x44,0x44,0x44,0x44,0x44,0x44,0x44,0x44	;61.-80. Pixel, blau
.db 0x77,0x77,0x77,0x77,0x77,0x77,0x77,0x77,0x77,0x77	;21.-100. Pixel, weiss
.db 0x33,0x33,0x33,0x33,0x33,0x33,0x33,0x33,0x33,0x33	;21.-120. Pixel, gelb
.db 0x55,0x55,0x55,0x55,0x55,0x55,0x55,0x55,0x55,0x55	;21.-140. Pixel, cyan
.db 0x66,0x66,0x66,0x66,0x66,0x66,0x66,0x66,0x66,0x66	;21.-160. Pixel, rotblau
.db 0x70,0x70,0x70,0x70,0x70,0x70,0x70,0x70,0x70,0x70	;21.-180. Pixel, grau
.db  0x00,0x00,0x00,0x00,0x00,0x0F			;181.-192. Pixel, schwarz
;; 126. Zeile:					
.db 0xF0,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00	;erste 10 Bytes = 20 Pixel, schwarz
.db 0x11,0x11,0x11,0x11,0x11,0x11,0x11,0x11,0x11,0x11	;21.-40. Pixel, rot
.db 0x22,0x22,0x22,0x22,0x22,0x22,0x22,0x22,0x22,0x22	;41.-60. Pixel, gruen
.db 0x44,0x44,0x44,0x44,0x44,0x44,0x44,0x44,0x44,0x44	;61.-80. Pixel, blau
.db 0x77,0x77,0x77,0x77,0x77,0x77,0x77,0x77,0x77,0x77	;21.-100. Pixel, weiss
.db 0x33,0x33,0x33,0x33,0x33,0x33,0x33,0x33,0x33,0x33	;21.-120. Pixel, gelb
.db 0x55,0x55,0x55,0x55,0x55,0x55,0x55,0x55,0x55,0x55	;21.-140. Pixel, cyan
.db 0x66,0x66,0x66,0x66,0x66,0x66,0x66,0x66,0x66,0x66	;21.-160. Pixel, rotblau
.db 0x70,0x70,0x70,0x70,0x70,0x70,0x70,0x70,0x70,0x70	;21.-180. Pixel, grau
.db  0x00,0x00,0x00,0x00,0x00,0x0F			;181.-192. Pixel, schwarz
;; 127. Zeile:					
.db 0xF0,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00	;erste 10 Bytes = 20 Pixel, schwarz
.db 0x11,0x11,0x11,0x11,0x11,0x11,0x11,0x11,0x11,0x11	;21.-40. Pixel, rot
.db 0x22,0x22,0x22,0x22,0x22,0x22,0x22,0x22,0x22,0x22	;41.-60. Pixel, gruen
.db 0x44,0x44,0x44,0x44,0x44,0x44,0x44,0x44,0x44,0x44	;61.-80. Pixel, blau
.db 0x77,0x77,0x77,0x77,0x77,0x77,0x77,0x77,0x77,0x77	;21.-100. Pixel, weiss
.db 0x33,0x33,0x33,0x33,0x33,0x33,0x33,0x33,0x33,0x33	;21.-120. Pixel, gelb
.db 0x55,0x55,0x55,0x55,0x55,0x55,0x55,0x55,0x55,0x55	;21.-140. Pixel, cyan
.db 0x66,0x66,0x66,0x66,0x66,0x66,0x66,0x66,0x66,0x66	;21.-160. Pixel, rotblau
.db 0x70,0x70,0x70,0x70,0x70,0x70,0x70,0x70,0x70,0x70	;21.-180. Pixel, grau
.db  0x00,0x00,0x00,0x00,0x00,0x0F			;181.-192. Pixel, schwarz
;; 128. Zeile:					
.db 0xF0,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00	;erste 10 Bytes = 20 Pixel, schwarz
.db 0x11,0x11,0x11,0x11,0x11,0x11,0x11,0x11,0x11,0x11	;21.-40. Pixel, rot
.db 0x22,0x22,0x22,0x22,0x22,0x22,0x22,0x22,0x22,0x22	;41.-60. Pixel, gruen
.db 0x44,0x44,0x44,0x44,0x44,0x44,0x44,0x44,0x44,0x44	;61.-80. Pixel, blau
.db 0x77,0x77,0x77,0x77,0x77,0x77,0x77,0x77,0x77,0x77	;21.-100. Pixel, weiss
.db 0x33,0x33,0x33,0x33,0x33,0x33,0x33,0x33,0x33,0x33	;21.-120. Pixel, gelb
.db 0x55,0x55,0x55,0x55,0x55,0x55,0x55,0x55,0x55,0x55	;21.-140. Pixel, cyan
.db 0x66,0x66,0x66,0x66,0x66,0x66,0x66,0x66,0x66,0x66	;21.-160. Pixel, rotblau
.db 0x70,0x70,0x70,0x70,0x70,0x70,0x70,0x70,0x70,0x70	;21.-180. Pixel, grau
.db  0x00,0x00,0x00,0x00,0x00,0x0F			;181.-192. Pixel, schwarz
;; 129. Zeile:					
.db 0xF0,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00	;erste 10 Bytes = 20 Pixel, schwarz
.db 0x11,0x11,0x11,0x11,0x11,0x11,0x11,0x11,0x11,0x11	;21.-40. Pixel, rot
.db 0x22,0x22,0x22,0x22,0x22,0x22,0x22,0x22,0x22,0x22	;41.-60. Pixel, gruen
.db 0x44,0x44,0x44,0x44,0x44,0x44,0x44,0x44,0x44,0x44	;61.-80. Pixel, blau
.db 0x77,0x77,0x77,0x77,0x77,0x77,0x77,0x77,0x77,0x77	;21.-100. Pixel, weiss
.db 0x33,0x33,0x33,0x33,0x33,0x33,0x33,0x33,0x33,0x33	;21.-120. Pixel, gelb
.db 0x55,0x55,0x55,0x55,0x55,0x55,0x55,0x55,0x55,0x55	;21.-140. Pixel, cyan
.db 0x66,0x66,0x66,0x66,0x66,0x66,0x66,0x66,0x66,0x66	;21.-160. Pixel, rotblau
.db 0x70,0x70,0x70,0x70,0x70,0x70,0x70,0x70,0x70,0x70	;21.-180. Pixel, grau
.db  0x00,0x00,0x00,0x00,0x00,0x0F			;181.-192. Pixel, schwarz
;; 130. Zeile:					
.db 0xF0,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00	;erste 10 Bytes = 20 Pixel, schwarz
.db 0x11,0x11,0x11,0x11,0x11,0x11,0x11,0x11,0x11,0x11	;21.-40. Pixel, rot
.db 0x22,0x22,0x22,0x22,0x22,0x22,0x22,0x22,0x22,0x22	;41.-60. Pixel, gruen
.db 0x44,0x44,0x44,0x44,0x44,0x44,0x44,0x44,0x44,0x44	;61.-80. Pixel, blau
.db 0x77,0x77,0x77,0x77,0x77,0x77,0x77,0x77,0x77,0x77	;21.-100. Pixel, weiss
.db 0x33,0x33,0x33,0x33,0x33,0x33,0x33,0x33,0x33,0x33	;21.-120. Pixel, gelb
.db 0x55,0x55,0x55,0x55,0x55,0x55,0x55,0x55,0x55,0x55	;21.-140. Pixel, cyan
.db 0x66,0x66,0x66,0x66,0x66,0x66,0x66,0x66,0x66,0x66	;21.-160. Pixel, rotblau
.db 0x70,0x70,0x70,0x70,0x70,0x70,0x70,0x70,0x70,0x70	;21.-180. Pixel, grau
.db  0x00,0x00,0x00,0x00,0x00,0x0F			;181.-192. Pixel, schwarz
;; 131. Zeile:
.db 0xF0,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00	;erste 10 Bytes = 20 Pixel, schwarz
.db 0x11,0x11,0x11,0x11,0x11,0x11,0x11,0x11,0x11,0x11	;21.-40. Pixel, rot
.db 0x22,0x22,0x22,0x22,0x22,0x22,0x22,0x22,0x22,0x22	;41.-60. Pixel, gruen
.db 0x44,0x44,0x44,0x44,0x44,0x44,0x44,0x44,0x44,0x44	;61.-80. Pixel, blau
.db 0x77,0x77,0x77,0x77,0x77,0x77,0x77,0x77,0x77,0x77	;21.-100. Pixel, weiss
.db 0x33,0x33,0x33,0x33,0x33,0x33,0x33,0x33,0x33,0x33	;21.-120. Pixel, gelb
.db 0x55,0x55,0x55,0x55,0x55,0x55,0x55,0x55,0x55,0x55	;21.-140. Pixel, cyan
.db 0x66,0x66,0x66,0x66,0x66,0x66,0x66,0x66,0x66,0x66	;21.-160. Pixel, rotblau
.db 0x70,0x70,0x70,0x70,0x70,0x70,0x70,0x70,0x70,0x70	;21.-180. Pixel, grau
.db  0x00,0x00,0x00,0x00,0x00,0x0F			;181.-192. Pixel, schwarz
;; 132. Zeile:					
.db 0xFF,0xFF,0xFF,0xFF,0xFF,0xFF,0xFF,0xFF,0xFF,0xFF	;erste 10 Bytes = 20 Pixel, weiss
.db 0xFF,0xFF,0xFF,0xFF,0xFF,0xFF,0xFF,0xFF,0xFF,0xFF	; - 40. Pixel, weiss
.db 0xFF,0xFF,0xFF,0xFF,0xFF,0xFF,0xFF,0xFF,0xFF,0xFF	; - 60. Pixel, weiss
.db 0xFF,0xFF,0xFF,0xFF,0xFF,0xFF,0xFF,0xFF,0xFF,0xFF	; - 80. Pixel, weiss
.db 0xFF,0xFF,0xFF,0xFF,0xFF,0xFF,0xFF,0xFF,0xFF,0xFF	; - 100. Pixel, weiss
.db 0xFF,0xFF,0xFF,0xFF,0xFF,0xFF,0xFF,0xFF,0xFF,0xFF	; - 120. Pixel, weiss
.db 0xFF,0xFF,0xFF,0xFF,0xFF,0xFF,0xFF,0xFF,0xFF,0xFF	; - 140. Pixel, weiss
.db 0xFF,0xFF,0xFF,0xFF,0xFF,0xFF,0xFF,0xFF,0xFF,0xFF	; - 160. Pixel, weiss
.db 0xFF,0xFF,0xFF,0xFF,0xFF,0xFF,0xFF,0xFF,0xFF,0xFF	; - 180. Pixel, weiss
.db 0xFF,0xFF,0xFF,0xFF,0xFF,0xFF			;181.-192. Pixel, weiss

#else

;; Bild mit groesserer Aufloesung: 240 breit, 264 hoch
;; 1. Zeile:					
.db 0xF0,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00	;erste 10 Bytes = 20 Pixel, schwarz
.db 0x11,0x11,0x11,0x11,0x11,0x11,0x11,0x11,0x11,0x11	;21.-40. Pixel, rot
.db 0x33,0x33,0x33,0x33,0x33,0x33,0x33,0x33,0x33,0x33	;21.-60. Pixel, gelb
.db 0x22,0x22,0x22,0x22,0x22,0x22,0x22,0x22,0x22,0x22	;41.-80. Pixel, gruen
.db 0x44,0x44,0x44,0x44,0x44,0x44,0x44,0x44,0x44,0x44	;61.-100. Pixel, blau
.db 0x77,0x77,0x77,0x77,0x77,0x77,0x77,0x77,0x77,0x77	;21.-120. Pixel, weiss
.db 0x70,0x70,0x70,0x70,0x70,0x70,0x70,0x70,0x70,0x70	;21.-140. Pixel, grau
.db 0x55,0x55,0x55,0x55,0x55,0x55,0x55,0x55,0x55,0x55	;21.-160. Pixel, cyan
.db 0x66,0x66,0x66,0x66,0x66,0x66,0x66,0x66,0x66,0x66	;21.-180. Pixel, rotblau
.db 0xF0,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00	;181.-200. Pixel, schwarz
.db 0xFF,0xFF,0xFF,0xFF,0xFF,0xFF,0xFF,0xFF,0xFF,0xFF	;181.-220. Pixel, weiss
.db 0xFF,0xFF,0xFF,0xFF,0xFF,0xFF,0xFF,0xFF,0xFF,0xFF	;181.-240. Pixel, weiss
;; da muessten noch mehr Zeilen kommen

#endif

;-----------------------------------------------------------------------------
.equ NormalerCompwert = (F_CPU*64/1000000)-1	 ;64 usec fuer Zeilenlaenge

#ifdef INTERLACE

#else

timer1_init:			;fuer Variante ohne Interlace
	ldi r16, (1<<WGM12)|(1<<CS10)	;CTC-OCR1A-Modus, Prescaler=1
	sts TCCR1B, r16			;beim Atmega1284P muss fuer einige io-Register sts statt out verwendet werden
	ldi r16, low(NormalerCompwert)
	ldi r17, high(NormalerCompwert)
	sts OCR1AH, r17
	sts OCR1AL, r16
	ldi r16, 0		;Startwert des Timers = 0
	sts TCNT1H, r16
	sts TCNT1L, r16
	ldi r16, (1<<OCIE1A)	;Timer1 Interrupts bei Vergleichswert
	sts TIMSK1, r16
	ldi zeilenzaehlerL, low(Z312-1)
	ldi zeilenzaehlerH, high(Z312-1) ;Starten mit letzter Zeile
	clr syncstatus
#ifdef HIRES
	ldi r16, 1		;nicht wirklich benoetigt bei HIRES
#else
	ldi r16, 2		;jeweils 2 Videozeilen gleich (bei LOWRES und MEDRES)
#endif
	mov r15, r16
	ret

#endif

;; Eine normale Video-Zeile sieht so aus:
;; 4.5us Sync-Puls, 7.5us Schwarzschulter, 0.75us Rand, 48.0us Text, 3.25us leer --> total 64us
;; 90 Takte         150 Takte              15           960 Takte    65 Takte           1280 Takte bei 20MHz

#ifdef INTERLACE

#else

timer1_match_A:			;Interrupt-Routine, Variante ohne INTERLACE
	out BILD_PORT, zero	;vordere Schwarzschulter 1.5us
	in r4, SREG
	push ZL
	push ZH
	jitterausgleich
	push r19
	push r16
	push r17
	push r18
	ticksw [VORDERE_SCHWARZSCHULTER-13-JITAKTE] ;Ticks von "in r4" bis und mit "cbi SYNC"
	cbi SYNC_PORT, SYNCBIT		;Synchronisationspuls starten (HSYNC oder VSYNC)
	adiw zeilenzaehlerL, 1
	tst syncstatus			;ist gesetzt wenn VSYNC ausgefuehrt wird
	brne t1_L2
	ldi r16, high(Z312)
	cpi zeilenzaehlerL, low(Z312)	;letzte Videozeile erreicht?
	cpc zeilenzaehlerH, r16		;7 Takte bis hier
	breq vsync_start
	clr r17			;r17 auf schwarz setzen
	ldi r16, high(oberedunkelzeilen)
	cpi zeilenzaehlerL, low(oberedunkelzeilen)
	cpc zeilenzaehlerH, r16		;12 Takte bis hier
	brcs leerzeile_1
	mov r17, hintergrundfarbe
	ldi r16, high(oberedunkelzeilen+RAND_OBEN)
	cpi zeilenzaehlerL, low(oberedunkelzeilen+RAND_OBEN)
	cpc zeilenzaehlerH, r16		;17 Takte bis hier
	brcs leerzeile_farbe_r17
	ldi r16, high(oberedunkelzeilen+RAND_OBEN+INHALTZEILEN)
	cpi zeilenzaehlerL, low(oberedunkelzeilen+RAND_OBEN+INHALTZEILEN)
	cpc zeilenzaehlerH, r16		;22 Takte bis hier
	brcs normale_zeile		;--> normale Videozeile darstellen (mit Bildinhalt)
leerzeile_1:			;14 Takte bis hier
	ticksw [5]
leerzeile_farbe_r17:		;19 Takte bis hier
	ticksw [SYNC1-21]	;inklusive 2 Takte fuer sbi abziehen
	sbi SYNC_PORT, SYNCBIT	;HSYNC Synchronisationspuls ende
	out BILD_PORT, r17
	rjmp t1_ret
vsync_start:
	ldi zeilenzaehlerL, 0		;Zeilenzaehler wieder auf Anfang, und VSYNC gestartet
	ldi zeilenzaehlerH, 0
	ldi ZL, low(testbild<<1)
	ldi ZH, high(testbild<<1)
	movw r2, ZL
	ldi r16, 2		;Statt hier 160 usec zu vertroedeln, 2 Interrupts abwarten und nur noch
	mov syncstatus, r16	;beim 3. mal die restlichen 160-2*64=32 usec abwarten
	rjmp t1_ret
t1_L2:
	dec syncstatus
	brne t1_ret
	rcall uhr_update	;+52 Takte --> 59 Takte bis hier
	ticksw [4*MU-59]	;4 Microsekunden bis hier
	ldi r16, SYNC3-2*64-4	;32 - Anzahl schon verbrauchte Microsekunden bis hier
	rcall usecwait		;160 Microsekunden fuer Seitensynchronisation (2*64+32)
	sbi SYNC_PORT, SYNCBIT	;VSYNC Synchronisationspuls ende
t1_ret:
	pop r18
	pop r17
	pop r16
	pop r19
	pop ZH
	pop ZL
	out SREG, r4
	reti
normale_zeile:			;20 Takte bis hier
;	rcall tickswait7	;Pause damit Verzoegerung gleich lang wie in Interlace-Variante
videozeile_darstellen:
	mov ZL, r2
	mov ZH, r3
	ticksw [SYNC1-22]
	sbi SYNC_PORT, SYNCBIT	;HSYNC Synchronisationspuls ende
	ticksw [HINTERE_SCHWARZSCHULTER]
	out BILD_PORT, hintergrundfarbe
	ticksw [RANDLINKS-8]	;Anzahl Takte bis zum ersten Bit abziehen
#ifdef MEDRES
	ldi r16, low(INHALT/5/2)	;wenn Voreingestellte Raender stimmen: 960/5=192, 192/2=96
					;also Bildbreite 192 Pixel
	rjmp _L1
_L0:	out BILD_PORT, r19
_L1:	lpm r19, Z+		;3 Takte
	nop			;1
	out BILD_PORT, r19	;1 --> zusammen 5
	swap r19		;1
	dec r16			;1
	brne _L0		;2 Takte --> zusammen mit out bei _L0: 5 Takte pro Pixel
	nop
	out BILD_PORT, r19	;1 Takt
	dec r15			;wird Widerhohlungszaehler gleich Null?
	brne _nochmals		;nein: gleiche Zeile nochmals
	ldi r16, 2
	mov r15, r16		;jeweils 2 Zeilen gleich
	out BILD_PORT, hintergrundfarbe
	rjmp _L3
_nochmals:
	movw ZL, r2
	out BILD_PORT, hintergrundfarbe
_L3:	movw r2, ZL
	pop r18
	pop r17
#endif ;//MEDRES

#ifdef HIRES
	ldi r16, low(INHALT/4/4)	;wenn Voreingestellte Raender stimmen: 960/16=60
	nop				;Bildbreite 960/4=240 Pixel
	rjmp _L1
_L0:	out BILD_PORT, r19	;4. Bit, 8., 12.
_L1:	lpm r19, Z+		;3 Takte
	out BILD_PORT, r19	;1. Bit ausgeben
	swap r19
	dec r16
	nop
	out BILD_PORT, r19	;2. Bit
	lpm r19, Z+
	out BILD_PORT, r19	;3. Bit
	swap r19
	brne _L0
	nop
	out BILD_PORT, r19
	movw r2, ZL
	pop r18
	out BILD_PORT, hintergrundfarbe
	pop r17
#endif ;//HIRES

	pop r16
	pop r19
	pop ZH
	pop ZL
	out SREG, r4
	reti

#endif ;//INTERLACE

;-----------------------------------------------------------------------------
; Hauptprogramm:
main:	ldi r16, low(RAMEND)
        out SPL, r16		; Init Stackpointer L
        ldi r16, high(RAMEND)
        out SPH, r16		; Init Stackpointer H
	clr zero		; r5 global mit 0 belegen
	ldi r16, (1<<BILDBITR)|(1<<BILDBITG)|(1<<BILDBITB)
	out BILD_DDR, r16	; Ausgang fuer Bildinhalt
	ldi r16, (1<<SYNCBIT)	; z.B. PC0
	out SYNC_DDR, r16	; Ausgang fuer Synchronisations-Pulse
	out SYNC_PORT, r16	; Ausgang gesetzt fuer inaktiv
	ldi hintergrundfarbe, HFARBE
	ldi textfarbe, TFARBE
	rcall uhr_reset
	clr altsec

#ifdef DEBUG
	sbi LED_DDR, LED1
	sbi LED_DDR, LED2
	sbi LED_PORT, LED1	;LED1 an
	sbi LED_PORT, LED2	;LED2 an
#endif

;	rcall uart_init
	rcall timer1_init
	sei			; Interrupts einschalten

;; Initialisierungen fuers Hauptprogramm:
	ldi status, 0		;r19=status

	ldi r16, (1<<SE)	;Schlafmodus IDLE setzen
	out SMCR, r16
#ifdef DEBUG
_L1:	sleep
	lds r16, uhr_sec
	cp r16, altsec
	breq _L1		;erste Sekunde abwarten
	mov altsec, r16
	cbi LED_PORT, LED1	;LED1 aus
	cbi LED_PORT, LED2	;LED2 aus
#endif
mainloop:
;	sleep			;nicht unbedingt noetig, koennte aber etwas Strom sparen
#ifdef DEBUG
	lds r16, uhr_sec
	cp r16, altsec
	breq mainloop		;naechste Sekunde abwarten
	mov altsec, r16
	cbi LED_PORT, LED1	;LED1 aus
	cbi LED_PORT, LED2	;LED2 aus
#endif
	rjmp mainloop

;-----------------------------------------------------------------------------
; Genaue Wartefunktionen:
; tickswait wartet genau soviele Takte wie in r16 steht, mindestens 18
; usecwait wartet genau soviele Microsekunden wie in r16 steht
;-----------------------------------------------------------------------------
tickswait:			;r16 = Anzahl zu wartende Takte (inklusive ldi und rcall und ret)
	cpi r16, 18		;Minimale Anzahl Ticks: bei weniger stimmts nicht
	brcs ti_ret
	subi r16, 14
	lsr r16
	brcs ti_L1
ti_L1:	lsr r16
	brcc ti_L3
	adiw r24, 0		;statt 2 mal nop
ti_L2:	nop
ti_L3:	dec r16
	brne ti_L2
ti_ret:	ret

tickswait17: nop
tickswait16: nop
tickswait15: nop
tickswait14: nop
tickswait13: nop
tickswait12: nop
tickswait11: nop
tickswait10: nop
tickswait9:  nop
tickswait8:  nop
tickswait7:  ret		;rcall und ret brauchen zusammen 7 Takte


#if (F_CPU==20000000)
us_L1:	adiw r24, 0
	adiw r24, 0
	adiw r24, 0
	adiw r24, 0
usecwait:			;r16 = Anzahl Microsekunden warten (bei 20MHz) (inklusive ldi und rcall und ret)
	adiw r24, 0
	adiw r24, 0
	adiw r24, 0
	adiw r24, 0
	nop
	dec r16
	brne us_L1
	nop
	ret
#endif

;-----------------------------------------------------------------------------
; divmodu16u8
;  16Bit vorzeichenlose Zahl dividieren durch 8Bit vorzeichenlosen Teiler
; Eingabe: r13:r12 = Zahl (16Bit), r18 = Teiler (8Bit)
; Ausgabe: r13:r12 = Resultat, r18 = Rest
; Speicherverbrauch: 34 Byte
; Anzahl Takte: 158 - 174
;  Wenn Teiler 0 ist wird das Resultat 0xFFFF sein
;-----------------------------------------------------------------------------
divmodu16u8:
	push r16
	push r17
	mov r16, r18
	clr r18
	ldi r17, 16
di_L1:	lsl r12
	rol r13
	rol r18
	cp r18, r16
	brcs di_L2
	sub r18, r16
	inc r12
di_L2:	dec r17
	brne di_L1
	pop r17
	pop r16
	ret

;-----------------------------------------------------------------------------
;; Uhr: Zeit seit dem Einschalten
uhr_reset:
	sts uhr_ticks, zero
	sts uhr_sec, zero
	sts uhr_min, zero
	sts uhr_std, zero
	sts uhr_tage, zero
	ret

uhr_update:			;Dieses Unterprogramm braucht immer genau die gleiche Anzahl Takte
	lds r16, uhr_ticks	;unabhaengig davon welche Daten wirklich aktualisiert werden.
	inc r16			;52 Takte inklusive rcall und ret,  r16-r17 zerstoert
	cpi r16, 50
	breq u_L1
	clr r17
	rjmp u_L1a
u_L1:	clr r16
	ldi r17, 1
u_L1a:	sts uhr_ticks, r16
	lds r16, uhr_sec
	add r16, r17
	cpi r16, 60
	breq u_L2
	clr r17
	rjmp u_L2a
u_L2:	clr r16
	ldi r17, 1
u_L2a:	sts uhr_sec, r16
	lds r16, uhr_min
	add r16, r17
	cpi r16, 60
	breq u_L3
	clr r17
	rjmp u_L3a
u_L3:	clr r16
	ldi r17, 1
u_L3a:	sts uhr_min, r16
	lds r16, uhr_std
	add r16, r17
	cpi r16, 24
	breq u_L4
	clr r17
	rjmp u_L4a
u_L4:	clr r16
	ldi r17, 1
u_L4a:	sts uhr_std, r16
	lds r16, uhr_tage
	add r16, r17
	sts uhr_tage, r16
	ret
